# flake8: noqa

from .main import *
from . import ops
from . import sim
from . import utils
from . import help
from . import demos
from .main import *

__docformat__ = 'restructuredtext'

#package-level docstring
__doc__ = """
Geonomics - A Python package for running complex landscape genomics simulations
===============================================================================

"""
#with open('../README.rst', 'r') as f:
#    __doc__ = __doc__ + f.read()




#!/usr/bin/python

#####################
#TODO

#Add a "timing" argument, that I can set to false to run the model without
#plotting, to get an accurate assessment of run time

#####################

# geonomics imports
from geonomics.utils.viz import _check_display

# other imports
import os
import numpy as np
import matplotlib as mpl
_check_display()
import matplotlib.pyplot as plt
from matplotlib import animation
from mpl_toolkits.axes_grid1 import make_axes_locatable
from mpl_toolkits.mplot3d.axes3d import Axes3D
import time
import rasterio as rio

try:
    from pykrige.ok import OrdinaryKriging
    with_pykrige = True
except ModuleNotFoundError:
    print(("NOTE: Module 'pykrige' not found. Yosemite demo 3d plots cannot be"
          " produced."))
    with_pykrige = False


gnx_dir = os.path.split(__file__)[0]
gnx_dir = os.path.join(*os.path.split(gnx_dir)[:-1])
DATA_PATH = os.path.join(gnx_dir, "data", "yosemite_demo")


def krig_phenotype(mod, window_width=8, rel_bandwidth=0.5, max_n_pts=10000):
    print('Kriging phenotype. Please wait...\n\n')
    if len(mod.comm[0]) > max_n_pts:
        inds = np.random.choice([*mod.comm[0].keys()], size=5000,
                                replace=False)
    else:
        inds = np.array([*mod.comm[0].keys()])
    xs = mod.comm[0]._get_x(inds)
    ys = mod.comm[0]._get_y(inds)
    zs = mod.comm[0]._get_z(individs=inds)[:,0]
    OK = OrdinaryKriging(xs, ys, zs, variogram_model='spherical')
    gridx = np.arange(0.5, mod.land.dim[0] + 0.5, 1)
    gridy = np.arange(0.5, mod.land.dim[1] + 0.5, 1)
    krig_pheno, ss = OK.execute('grid', gridx, gridy)
    return krig_pheno


# plot 3 change rasters, 1 for each of the 3 landscape layers
def make_change_fig(mod):
    change_fig = plt.figure()
    change_fig_titles = {0: 'temperature',
                         2: 'precipitation',
                         1: 'habitat suitability',
                     }
    cmaps = ['afmhot_r', 'rainbow_r', 'PiYG']
    for i, d in enumerate(change_fig_titles.items()):
        ax = change_fig.add_subplot(1, 3, i + 1)
        lyr_num = d[0]
        title = d[1]
        ax.set_title(title, size=20)
        af = mod.land[lyr_num]._get_rast_in_native_units()
        b4 = mod.orig_land[lyr_num]._get_rast_in_native_units()
        diff = af - b4
        im = ax.imshow(diff, cmap = cmaps[i])
        units = '$\Delta$ ' + mod.land[lyr_num].units
        cb = change_fig.colorbar(im, orientation='horizontal')
        cb.set_label(label=units, size=13)
        cb.ax.tick_params(labelsize=9)
        (xtick_locs, xtick_labs, ytick_locs,
         ytick_labs) = mod.land[lyr_num]._get_coord_ticks()
        ax.set_xticks(xtick_locs)
        ax.set_xticklabels(np.round(xtick_labs, 1), rotation=45, size=7,
                           color='gray')
        ax.set_yticks(ytick_locs)
        ax.set_yticklabels(ytick_labs, size=7, color='gray')
        ax.set_xlabel('lon', color='gray', size=10)
        if i == 0:
            ax.set_ylabel('lat', color='gray', size=10)
    change_fig.show()
    return change_fig


# drape a raster on top of a DEM
def drape_raster(mod, rast, DEM, cbar_label, gif_filename,
                 save_figs, make_gifs, cmap='rainbow'):
    fig3d = plt.figure()
    ax3d = fig3d.add_subplot(111, projection='3d')
    ax3d.set_xlabel('lon', color='gray', size=10)
    ax3d.set_ylabel('lat', color='gray', size=10)
    ax3d.set_zlabel('alt (m)', color='gray', size=10)
    # get the x and y coordinates for each cell
    xi = range(mod.land.dim[0])
    yi = range(mod.land.dim[1])
    # use the x and y coords to get a meshgrid
    X, Y = np.meshgrid(xi, yi)
    # create a mappable object, to use to make the colorbar
    mappable = plt.cm.ScalarMappable()
    mappable.set_array(rast)
    mappable.set_cmap(cmap)

    # initialization function, which plots the background of each frame
    def init():
        # plot the DEM as a surface, coloring its patches by the draped raster
        surf = ax3d.plot_surface(X, Y, DEM, facecolors=mappable.cmap(rast),
                                 cmap=mappable.cmap)
        # add a 2d contour plot projected onto the horizontal below it
        cont = ax.contourf(X, Y, rast, zdir='elevation', offset=np.min(DEM),
                           cmap=mappable.cmap)
        # create a colorbar
        cb = plt.colorbar(mappable)
        cb.set_label(label=cbar_label, size=13)
        cb.ax.tick_params(labelsize=9)
        # set the x-, y-, and z-axis ticks and labels
        (xtlocs, xtlabs, ytlocs, ytlabs) = mod.land[0]._get_coord_ticks()
        ztlocs = ax.zaxis.get_ticklocs()
        ztlabs = [int(txt.get_text()) for txt in ax.zaxis.get_ticklabels()]
        ax.set_xticks(xtlocs)
        ax.set_xticklabels(np.round(xtlabs, 1), size=7, color='gray')
        ax.set_yticks(ytlocs)
        ax.set_yticklabels(np.round(ytlabs, 1), size=7, color='gray')
        ax.set_zticks(ztlocs)
        ax.set_zticklabels(np.round(ztlabs, 0), size=7, color='gray')

        return fig3d,

    if make_gifs:
        # animation function, which will be called sequentially
        def animate(i):
            ax3d.view_init(elev=5., azim=i)
            return fig3d,
        anim = animation.FuncAnimation(fig3d, animate, init_func=init, frames=359,
                                       interval=5, blit=True)
    else:
        ax3d.view_init(elev=5., azim=115)
        init()

    if save_figs:
        try:
            anim.save(gif_filename, writer='imagemagick', fps=60)
        except Exception as e:
            print(('\nCould not use Imagemagick to save the 3D plot\'s '
                   'GIF.  The following error was thrown:\n\t%s') % e)

    fig3d.show()

    return fig3d


# create and save images that Imagemagick will stitch into a GIF
def save_model_gif(mod, save_figs, make_gifs, n_individs=5000):
    # set up second figure, for a GIF animation
    mid_col_width = 0.8
    scnd_col_width = 1
    gif_fig = plt.figure(figsize=(6.75 + mid_col_width * 6.75, 5.4))
    ax2_gs = gif_fig.add_gridspec(1, 3, width_ratios=[1, scnd_col_width, 1])
    # gif_fig, axs2 = plt.subplots(1, 2)
    # gif_fig.set_tight_layout(True)
    ax2_1 = gif_fig.add_subplot(ax2_gs[0, 0])
    ax2_2 = gif_fig.add_subplot(ax2_gs[0, 2])
    gif_fig.suptitle('Time step: %i' % (mod.t+1), size=20)
    ax2_1.set_title('Temperature', size=12)
    ax2_2.set_title('Habitat suitability', size=12)
    ax2_1.set_xlabel('lon', size=8)
    ax2_1.set_ylabel('lat', size=8)
    ax2_2.set_xlabel('lon', size=8)
    ax2_2.set_ylabel('lat', size=8)
    ax2_1.set_aspect(1)
    ax2_2.set_aspect(1)
    # plot the population on both the temp and hab rasters
    land1 = ax2_1.pcolormesh(mod.land._x_cell_bds,
                             mod.land._y_cell_bds,
                             mod.land[0]._get_rast_in_native_units(),
                             cmap='coolwarm',
                             vmin=mod.land[0]._scale_min,
                             vmax=mod.land[0]._scale_max)
    xticks, xticklabs, yticks, yticklabs = mod.land[0]._get_coord_ticks()
    ax2_1.set_xticklabels(labels=xticklabs, size=6)
    ax2_1.set_yticklabels(labels=yticklabs, size=6)
    divider = make_axes_locatable(ax2_1)
    cax = divider.append_axes("right", size="5%", pad=0.05)
    cbar_ticks = mod.land[0]._get_cbar_ticks_and_minmax_scaled_vals()
    cbar = plt.colorbar(land1, cax=cax, ticks=np.linspace(cbar_ticks[1],
                                                          cbar_ticks[2], 5))
    cbar.ax.set_yticklabels(cbar_ticks[0], size=6)
    cbar.set_label(mod.land[0].units, rotation=270, labelpad=5, y=0.5,
                   size=7)
    # get a random subset of n_individs individuals, to use in the plot
    individs = mod.comm[0]._get_random_individuals(n_individs)
    coords = mod.comm[0]._get_plot_coords(individs=individs)
    xs = coords[:, 0]
    ys = coords[:, 1]
    zs = mod.comm[0]._get_z(individs=individs)[:, 0]
    ax2_1.scatter(xs, ys, c=zs, s=1, cmap='coolwarm')
    land2 = ax2_2.pcolormesh(mod.land._x_cell_bds,
                             mod.land._y_cell_bds,
                             mod.land[1]._get_rast_in_native_units(),
                             cmap='BrBG_r',
                             vmin=mod.land[1]._scale_min,
                             vmax=mod.land[1]._scale_max)
    xticks, xticklabs, yticks, yticklabs = mod.land[1]._get_coord_ticks()
    ax2_2.set_xticklabels(labels=xticklabs, size=6)
    ax2_2.set_yticklabels(labels=yticklabs, size=6)
    divider = make_axes_locatable(ax2_2)
    cax = divider.append_axes("right", size="5%", pad=0.05)
    cbar_ticks = mod.land[1]._get_cbar_ticks_and_minmax_scaled_vals()
    cbar = plt.colorbar(land2, cax=cax, ticks=np.linspace(cbar_ticks[1],
                                                          cbar_ticks[2], 5))
    cbar.ax.set_yticklabels(cbar_ticks[0], size=6)
    cbar.set_label(mod.land[1].units, rotation=270, labelpad=15, y=0.5,
                   size=10)
    ax2_2.scatter(xs, ys, c='black', s=0.5)

    # save the figure
    if save_figs and make_gifs:
        gif_fig.savefig('gif_img_%i.jpg' % (mod.t + 1))

    # manually close the figure
    plt.close(gif_fig)



def _make_params():

    # This is a parameters file generated by Geonomics
    # (by the gnx.make_parameters_file() function).


                       ##  ::::::          :::    :: ::::::::::##
                 ##:::::    ::::   :::      ::    :: :: ::::::::::: :##
              ##::::::::     ::            ::   ::::::::::::::::::::::::##
            ##:::::::::                      :::::::::: :::::: ::::::::  :##
          ## : ::::  ::                    ::::  : ::    :::::::: : ::  :   ##
         ##GGGGG  EEEEE OOOOO   NN   NN   OOOOO   MM   MM IIIIII  CCCCC SSSSS##
        ##GG     EE    OO   OO  NNN  NN  OO   OO  MM   MM   II   CC     SS    ##
        ##GG     EE   OO     OO NN N NN OO     OO MMM MMM   II   CC     SSSSSS##
        ##GG GGG EEEE OO     OO NN  NNN OO     OO MM M MM   II   CC         SS##
        ##GG   G EE    OO   OO  NN   NN  OO   OO  MM   MM   II   CC        SSS##
         ##GGGGG  EEEEE OOOOO   NN   NN   OOOOO   MM   MM IIIIII  CCCCC SSSSS##
          ##    :::::::::               :::::::::: ::              ::  :   :##
            ##:   :::::                    :::::: :::             ::::::: ##
              ##   :::                      :::::  ::              :::::##
                 ## ::                      ::::                     ##
                       ##                                      ##
                          ## :: ::    :::            ##


    params = {
    ###############################################################################

    ###################
    #### LANDSCAPE ####
    ###################
        'landscape': {

        ##############
        #### main ####
        ##############
            'main': {

                # NOTE: The resolution of the rasters is 0.00833 dec. deg.,
                # which is equal to about 0.00833* 88070.4 = 730.984 m in E-W
                # and about 0.00833 * 111320 = 927.2956 m in N-S directions.
                # NOTE: Thus, a cell is about 730.984 * 927.2956 = 678000 m^2,
                # which is = 67.8 hectares.
                # NOTE: The difference in distance in E-W and N-S is problematic
                # for distance-based things!!!!

                #y,x (a.k.a. i,j) dimensions of the Landscape
                'dim':                      (90, 90),
                #resolution of the Landscape
                'res':                      (0, 0),
                #upper-left corner of the Landscape
                'ulc':                      (0, 0),
                #projection of the Landscape
                'prj':                      None,
                }, # <END> 'main'

        ################
        #### layers ####
        ################
            'layers': {

                #layer name (LAYER NAMES MUST BE UNIQUE!)
                'tmp': {

            #######################################
            #### layer num. 0: init parameters ####
            #######################################

                    #initiating parameters for this layer
                    'init': {

                        #parameters for a 'file'-type Layer
                        'file': {
                            #</path/to/file>.<ext>
                            'filepath': os.path.join(DATA_PATH,
                                                     'yosemite_lyrs',
                                                     'tmp_1980-2010_90x90.tif'),
                            #minimum value to use to rescale the Layer to [0,1]
                            'scale_min_val':                -1.37,
                            #maximum value to use to rescale the Layer to [0,1]
                            'scale_max_val':                19.75773,
                            #decimal precision to use for coord-units (ulc & res)
                            'coord_prec':                   8,
                            #units of this file's variable
                            'units':                    '$^∘C$',

                            }, # <END> 'file'

                        }, # <END> 'init'

                #########################################
                #### layer num. 0: change parameters ####
                #########################################

                    #landscape-change events for this Layer
                    'change': {

                        0: {
                            #array or file for final raster of event, or directory
                            #of files for each stepwise change in event
                            'change_rast': os.path.join(DATA_PATH,
                                                        'yosemite_lyrs',
                                                        'tmp'),

                            #starting timestep of event
                            'start_t':          509,
                            #ending timestep of event
                            'end_t':            594,
                            #number of stepwise changes in event
                            'n_steps':          18,
                            }, # <END> event 0

                        }, # <END> 'change'

                    }, # <END> layer num. 0


                #layer name (LAYER NAMES MUST BE UNIQUE!)
                'hab': {

            #######################################
            #### layer num. 1: init parameters ####
            #######################################

                    #initiating parameters for this layer
                    'init': {

                        #parameters for a 'file'-type Layer
                        'file': {
                            #</path/to/file>.<ext>
                            'filepath': os.path.join(DATA_PATH,
                                                     'yosemite_lyrs',
                                                     'sdm_1980-2010_90x90.tif'),
                            #minimum value to use to rescale the Layer to [0,1]
                            'scale_min_val':                0,
                            #maximum value to use to rescale the Layer to [0,1]
                            'scale_max_val':                1,
                            #decimal precision to use for coord-units (ulc & res)
                            'coord_prec':                   8,
                            #units of this file's variable
                            'units':                    'suitability',

                            }, # <END> 'file'

                        }, # <END> 'init'

                #########################################
                #### layer num. 1: change parameters ####
                #########################################

                    #landscape-change events for this Layer
                    'change': {

                        0: {
                            #array or file for final raster of event, or directory
                            #of files for each stepwise change in event
                            'change_rast': os.path.join(DATA_PATH,
                                                        'yosemite_lyrs',
                                                        'sdm'),

                            #starting timestep of event
                            'start_t':          509,
                            #ending timestep of event
                            'end_t':            594,
                            #number of stepwise changes in event
                            'n_steps':          18,
                            }, # <END> event 0

                        }, # <END> 'change'

                    }, # <END> layer num. 1


                #layer name (LAYER NAMES MUST BE UNIQUE!)
                'ppt': {

            #######################################
            #### layer num. 2: init parameters ####
            #######################################

                    #initiating parameters for this layer
                    'init': {

                        #parameters for a 'file'-type Layer
                        'file': {
                            #</path/to/file>.<ext>
                            'filepath': os.path.join(DATA_PATH,
                                                     'yosemite_lyrs',
                                                     'ppt_1980-2010_90x90.tif'),
                            #minimum value to use to rescale the Layer to [0,1]
                            'scale_min_val':                81.53713,
                            #maximum value to use to rescale the Layer to [0,1]
                            'scale_max_val':                2171.341,
                            #decimal precision to use for coord-units (ulc & res)
                            'coord_prec':                   8,
                            #units of this file's variable
                            'units':                    '$mm/yr$',

                            }, # <END> 'file'

                        }, # <END> 'init'

                #########################################
                #### layer num. 0: change parameters ####
                #########################################

                    #landscape-change events for this Layer
                    'change': {

                        0: {
                            #array or file for final raster of event, or directory
                            #of files for each stepwise change in event
                            'change_rast': os.path.join(DATA_PATH,
                                                        'yosemite_lyrs',
                                                        'ppt'),

                            #starting timestep of event
                            'start_t':          509,
                            #ending timestep of event
                            'end_t':            594,
                            #number of stepwise changes in event
                            'n_steps':          18,
                            }, # <END> event 0

                        }, # <END> 'change'

                    }, # <END> layer num. 0



        #### NOTE: Individual Layers' sections can be copy-and-pasted (and
        #### assigned distinct keys and names), to create additional Layers.


                } # <END> 'layers'

            }, # <END> 'landscape'


    ###############################################################################

    ###################
    #### COMMUNITY ####
    ###################
        'comm': {

            'species': {

                #species name (SPECIES NAMES MUST BE UNIQUE!)
                'Sceloporus graciosus': {

                #####################################
                #### spp num. 0: init parameters ####
                #####################################

                    'init': {
                        #starting number of individs
                        'N':                50000,
                        #carrying-capacity Layer name
                        'K_layer':          'hab',
                        #multiplicative factor for carrying-capacity layer
                        # NOTE: each cell is ~53.43 hectares, and pop density
                        # estimates in Utah (closest estimate I could find to
                        # Yosemite region) were 208/hectare, with little variation
                        # (Tinkle 1973).
                        # That would suggest each cell should have a carrying
                        # capacity of ~ 53.43 * 208. However, in the Sierra they 
                        # and S. occidentalis segregate by habitat (Rose 1976),
                        # so if we assume that only about 10% of land area
                        # is covered by S. graciosus' preferred open,
                        # exposed habitat, then that suggests we should
                        # use a K_factor of 67.8 * 208 * 0.1 = 1410.24
                        # NOTE: dividing by 10 for tractability
                        'K_factor':         1410.24 / 10,
                        }, # <END> 'init'

                #######################################
                #### spp num. 0: mating parameters ####
                #######################################

                    'mating'    : {
                        #age(s) at sexual maturity (if tuple, female first)
                        # NOTE: average reproductive age is 2 (i.e. in the third
                        # season) (Tinkle 1973; Tinkle et. al 1993)
                        'repro_age':                2,
                        #whether to assign sexes
                        'sex':                      True,
                        # NOTE: Some studies suggest a skewed sex ratio
                        # becuase of uneven male/female survival rate ratio
                        # (e.g. Tinkle 1973), but other, more recent studies do not
                        # (e.g. Tinkle et al. 1993), so just leaving sex_ratio at 1
                        #ratio of males to females
                        'sex_ratio':                1/1,
                        # NOTE: found no information about this, but 0.5 seems
                        # a reasonable default value
                        #intrinsic growth rate
                        'R':                        0.5,
                        #intrinsic birth rate (MUST BE 0<=b<=1)
                        'b':                        1,
                        #expectation of distr of n offspring per mating pair
                        #NOTE: using a lambda of equal to
                        # avg_clutch_size * avg_n_clutches_per_yr * avg_surv_rate_to_yr_1
                        # which I put at 4.464 * 2 * 0.16 = 1.428
                        # (where the first value is avg of all CA pops in Tinkle et al 1993,
                        # second value is from Tinkle 1973 and Tinkel et al 1993, and
                        # third comes from the Mt. Diablo population in Ruth 1978,
                        #cited in Tinkle et al. 1993)
                        'n_births_distr_lambda':    1.428,
                        #whether n births should be fixed at n_births_dist_lambda
                        'n_births_fixed':           False,
                        #radius of mate-search area
                        #NOTE: just have individuals mate close by (i.e. within
                        #12.457m, the average interannual movement distance
                        'mating_radius':            0.5,
                        'choose_nearest_mate':      False,
                        'inverse_dist_mating':      False,
                        }, # <END> 'mating'

                ##########################################
                #### spp num. 0: mortality parameters ####
                ##########################################

                    'mortality'     : {
                        #maximum age
                        # NOTE: 8 is the max age mentioned in Stebbins 1948's
                        # Lassen population study
                        'max_age':                      8,
                        #min P(death) (MUST BE 0<=d_min<=1)
                        'd_min':                        0,
                        #max P(death) (MUST BE 0<=d_max<=1)
                        'd_max':                        1,
                        #width of window used to estimate local pop density
                        'density_grid_window_width':    None,
                        }, # <END> 'mortality'

                #########################################
                #### spp num. 0: movement parameters ####
                #########################################

                    'movement': {
                        #NOTE: the movement and dispersal Wald distributions
                        # are parameterized based on the finding of
                        # Stebbins 1948 that these lizards have extremely
                        # small home ranges;
                        # expressed in cell-widths (where the cell width is
                        # ~730.984 m, according to raster res converted
                        # from dec. deg to m; see NOTE at top of file);
                        # so becuase the mean dispersal distance of
                        # Stebbins' lizards in that study is 12.457m (where
                        # "within/at edge of home area was treated as 0),
                        # this would be 0.01704 cell widths;
                        # then I assigned a sigma that allows for occasional
                        # longer-distance movement events,
                        # but not for long-distance offspring dispersal events

                        #whether or not the species is mobile
                        'move':                     True,
                        #mode of distr of movement direction
                        'direction_distr_mu':       1,
                        #concentration of distr of movement direction
                        'direction_distr_kappa':    0,
                        #mean of distr of movement distance
                        'movement_distance_distr_param1':        7e-5,
                        #variance of distr of movement distance
                        'movement_distance_distr_param2':     0.3,
                        #mean of distr of dispersal distance
                        'movement_distance_distr':      'lognormal',
                        'dispersal_distance_distr_param1':       7.5e-4,
                        #variance of distr of dispersal distance
                        'dispersal_distance_distr_param2':    1,
                        'dispersal_distance_distr':         'lognormal',
                        'move_surf'     : {
                            #move-surf Layer name
                            'layer':                'hab',
                            #whether to use mixture distrs
                            'mixture':              True,
                            #concentration of distrs
                            'vm_distr_kappa':       12,
                            #length of approximation vectors for distrs
                            'approx_len':           5000,
                            }, # <END> 'move_surf'

                        },    # <END> 'movement'


                #####################################################
                #### spp num. 0: genomic architecture parameters ####
                #####################################################

                    'gen_arch': {
                        #file defining custom genomic arch
                        'gen_arch_file':            None,
                        #num of loci
                        'L':                        1000,
                        #starting allele freq (None to draw freqs randomly)
                        #at 0.5
                        'start_p_fixed':            0.5,
                        #genome-wide per-base neutral mut rate (0 to disable)
                        'mu_neut':                  0,
                        #genome-wide per-base deleterious mut rate
                        #(0 to disable)
                        'mu_delet':                 0,
                        #shape of distr of deleterious effect sizes
                        'delet_alpha_distr_shape':  0.01,
                        #scale of distr of deleterious effect sizes
                        'delet_alpha_distr_scale':  100,
                        #alpha of distr of recomb rates
                        'r_distr_alpha':            None,
                        #beta of distr of recomb rates
                        'r_distr_beta':             None,
                        #whether loci should be dominant (for allele '1')
                        'dom':                      False,
                        #whether to allow pleiotropy
                        'pleiotropy':               False,
                        #custom fn for drawing recomb rates
                        'recomb_rate_custom_fn':    None,
                        #number of recomb paths to hold in memory
                        'n_recomb_paths_mem':       int(1e4),
                        #total number of recomb paths to simulate
                        'n_recomb_paths_tot':       int(1e5),
                        'n_recomb_sims':            10000,
                        'start_neut_zero':          False,
                        'allow_ad_hoc_recomb':      False,
                        'jitter_breakpoints':       False,
                        #whether to save mutation logs
                        'mut_log':                  False,

                        'traits': {

                            ###########################
                            ####trait 0 parameters ####
                            ###########################
                            #trait name (TRAIT NAMES MUST BE UNIQUE!)
                            'thermal_tol': {
                                #trait-selection Layer name
                                'layer':                'tmp',
                                #polygenic selection coefficient
                                'phi':                  0.5,
                                #number of loci underlying trait
                                'n_loci':               100,
                                #mutation rate at loci underlying trait
                                'mu':                   0,
                                #mean of distr of effect sizes
                                'alpha_distr_mu' :      0.1,
                                #variance of distr of effect size
                                'alpha_distr_sigma':    0,
                                #max allowed magnitude for an alpha value
                                'max_alpha_mag':        None,
                                #curvature of fitness function
                                'gamma':                1,
                                #whether the trait is universally advantageous
                                'univ_adv':             False
                                }, # <END> trait 0


        #### NOTE: Individual Traits' sections can be copy-and-pasted (and
        #### assigned distinct keys and names), to create additional Traits.


                            }, # <END> 'traits'

                        }, # <END> 'gen_arch'


                    }, # <END> spp num. 0



        #### NOTE: individual Species' sections can be copy-and-pasted (and
        #### assigned distinct keys and names), to create additional Species.


                }, # <END> 'species'

            }, # <END> 'comm'


    ##########################################################################

    ###############
    #### MODEL ####
    ###############
        'model': {
            #total Model runtime (in timesteps)
            'T':            600,
            #min burn-in runtime (in timesteps)
            'burn_T':       50,
            #seed number
            'num':          None,
            #time step interval for simplification of tskit tables
            'tskit_simp_interval':      100,

            ###############################
            #### iterations parameters ####
            ###############################
            'its': {
                #num iterations
                'n_its':            1,
                #whether to randomize Landscape each iteration
                'rand_landscape':   False,
                #whether to randomize Community each iteration
                'rand_comm':        False,
                #whether to burn in each iteration
                'repeat_burn':      False,
                }, # <END> 'iterations'


            ####################################
            #### data-collection parameters ####
            ####################################
            'data': {
                'sampling': {
                    #sampling scheme {'all', 'random', 'point', 'transect'}
                    'scheme':               'transect',
                    #sample size at each point, for point & transect sampling
                    'n':                    250,
                    #coords of collection points, for point sampling
                    'points':               None,
                    #coords of transect endpoints, for transect sampling
                    'transect_endpoints':   [(25, 25), (75, 5)],
                    #num points along transect, for transect sampling
                    'n_transect_points':    6,
                    #collection radius around points, for point & transect
                    #sampling
                    'radius':               10,
                    #when to collect data
                    'when':                 100,
                    #whether to save current Layers when data is collected
                    'include_landscape':    False,
                    #whether to include fixed loci in VCF files
                    'include_fixed_sites':  True,
                    },
                'format': {
                    #format for genetic data {'vcf', 'fasta'}
                    'gen_format':           ['vcf', 'fasta'],
                    #format for vector geodata {'csv', 'shapefile', 'geojson'}
                    'geo_vect_format':      'csv',
                    #format for raster geodata {'geotiff', 'txt'}
                    'geo_rast_format':      'geotiff',
                    },
                }, #<END> 'data'


            } # <END> 'model'

        } # <END> params

    return params


def _run(params, save_figs=False, time_it=False, make_gifs=False,
         make_3d_plots=False, n_individs_plot=10_000):
    params.model['name'] = 'yosemite_demo'
    # set the amount of time before and after climate change
    t_before_cc = 500
    t_after_cc = 100

    # create colormap to match the phenotype colors
    z_cmap = mpl.cm.coolwarm

    # start timer
    if time_it:
        start = time.time()

    # create the model
    from .. import make_model
    mod = make_model(params, verbose=True)

    # set plotting params
    ms = 6
    ax_fontdict = {'fontsize': 12,
                   'name': 'Bitstream Vera Sans'}
    ttl_fontdict = {'fontsize': 15,
                    'name': 'Bitstream Vera Sans'}

    # burn in, then plot starting population, on both rasters
    mod.walk(20000, 'burn')

    # run for the first 500 timsteps, before climate change
    for _ in range(t_before_cc):
        mod.walk(1)
        if save_figs and make_gifs:
            save_model_gif(mod, save_figs, make_gifs,
                           n_individs=n_individs_plot)

    # calculate neigh-mean raster before climate-change
    if with_pykrige:
        neigh_mean_b4 = krig_phenotype(mod, save_figs, make_gifs)
        try:
            np.savetxt('b4_cc_krig.txt', neigh_mean_b4)
        except Exception as e:
            print(('COULD NOT SAVE KRIGED PHENOTYPE BEFORE CLIMATE CHANGE:'
                  ' %s') % e)

    # walk for 100 more timesteps, then plot again,
    # at end of climate-change period
    for _ in range(t_after_cc):
        mod.walk(1)
        if save_figs and make_gifs:
            save_model_gif(mod, save_figs, make_gifs,
                           n_individs=n_individs_plot)

    # end timer
    if time_it:
        stop = time.time()
        tot_time = stop-start

    # calculate neigh-mean raster after climate-change
    if with_pykrige:
        neigh_mean_af = krig_phenotype(mod)
        try:
            np.savetxt('af_cc_krig.txt', neigh_mean_af)
        except Exception as e:
            print(('COULD NOT SAVE KRIGED PHENOTYPE AFTER CLIMATE CHANGE:'
                  ' %s') % e)


        # calculate the difference between the two neigh-mean rasters
        neigh_mean_diff = neigh_mean_af - neigh_mean_b4

    # create and save a population-size plot
    mod.walk(50)
    Nt_fig = plt.figure()
    Nt_ax = Nt_fig.add_subplot(111)
    #burn_len = mod.burn_t
    line_height = int(10000*np.ceil(max(mod.comm[0].Nt[-mod.t:])/10000))
    Nt_ax.plot(range(len(mod.comm[0].Nt[-mod.t:])), mod.comm[0].Nt[-mod.t:])
    #Nt_ax.plot([burn_len, burn_len], [0, line_height], c='red')
    Nt_ax.plot([500, 500], [0, line_height], c='red')
    Nt_ax.plot([600, 600], [0, line_height], c='red')
    chng_lyr_path = os.path.join(DATA_PATH, 'yosemite_lyrs/ppt/')
    chng_yrs = [int(x[:3]) for x in os.listdir(chng_lyr_path)]
    for yr in chng_yrs:
        Nt_ax.plot([yr, yr], [0, line_height], ':r',
                   linewidth=0.5)
    Nt_ax.set_label('time (time steps/years)')
    Nt_ax.set_ylabel('total population size (individuals)')
    if save_figs:
        Nt_fig.savefig('yosemite_time_series_pop_size.png', format='png',
                       dpi=1000)

    # make the change-raster plot
    change_fig = make_change_fig(mod)
    if save_figs:
        change_fig.savefig('yosemite_change_fig.png', format='png', dpi=1000)

    # create the GIF using imagemagick
    if make_gifs:
        try:
            os.system('cd %s' % gif_dir)
            os.system(('cd %s; convert -delay 5 -loop 0 `ls -v` '
                       '../yosemite.gif; cd ..') % gif_dir)
            os.system('..')
        except Exception as e:
            print(('\nCould not use Imagemagick to create the GIF. '
                   'The following error was thrown:\n\t%s') % e)

    # make the draped-raster plot
    if with_pykrige:
        DEM = rio.open(os.path.join(DATA_PATH,
                                    'yosemite_DEM_90x90.tif')).read()[0, :, :]
        if make_3d_plots:
            plt.rc('animation', html='html5')
            pheno_drape_fig = drape_raster(mod, neigh_mean_b4, DEM, 'phenotype',
                                           'yosemite_pheno_drape_fig.gif',
                                           save_figs, make_gifs,
                                           cmap='coolwarm')
        #pheno_chng_cmap = mpl.colors.LinearSegmentedColormap.from_list(
        #    'custom', [(0, plt.cm.coolwarm(0.075)), (1,
        #                                             plt.cm.coolwarm(0.925))])
            change_drape_fig = drape_raster(mod, neigh_mean_diff, DEM,
                                            'change in phenotype',
                                            'yosemite_pheno_change_drape_fig.gif',
                                            save_figs, make_gifs, cmap='PuOr_r')
            if save_figs:
                pheno_drape_fig.savefig('yosemite_pheno_drape_fig.png',
                                        format='png', dpi=1000)
                change_drape_fig.savefig('yosemite_pheno_change_drape_fig.png',
                                         format='png', dpi=1000)


    # print out time
    if time_it:
        print("\n\nModel ran in %0.2f seconds." % tot_time)

    return mod




#!/usr/bin/python
#data.py


'''
Classes and functions for implementing data collection and output
'''

#geonmics imports
from geonomics.utils.io import (_write_csv, _write_shapefile, _write_geojson,
                                _write_file)

#other imports
import numpy as np
from random import sample as rand_sample
from numpy import random as r
import os, sys
import datetime
import re
from shapely.geometry import Point
from shapely.ops import cascaded_union
import pandas as pd
import geopandas as gpd
from itertools import chain


######################################
# -----------------------------------#
# CLASSES ---------------------------#
# -----------------------------------#
######################################

#a DataCollector class, to paramtereize and manage data 
#collection, and write data to disk
class _DataCollector:
    def __init__(self, model_name, params):

    #some lookup dicts for writing data 
        self.file_extension_dict =   {'vcf': 'vcf',
                            'fasta': 'fasta',
                            'csv': 'csv',
                            'shapefile': 'shp',
                            'geojson': 'json',
                            'geotiff': 'tif'
                            }

        self.write_geodata_fn_dict = {'csv': _write_csv,
                            'shapefile': _write_shapefile,
                            'geojson': _write_geojson,
                            }

        #set model name and T
        self.model_name = model_name
        self.T = params.model.T

        #grab the params['data'] contents into objects
        sampling_params = params.model.data.sampling
        format_params = params.model.data.format

        #get the sampling scheme
        self.scheme = sampling_params.scheme

        #and run some asserts
        assert self.scheme in ['all', 'random',
            'point', 'transect'], ("The sampling scheme provided in the "
            "parameters must be one of the following values: 'all', 'random', "
            "'point', or 'transect'.")

        if sampling_params.scheme != 'all':
            assert 'n' in sampling_params.keys(), ("If the "
            "sampling scheme is not 'all' then the 'n' parameter must be "
            "defined, indicating the number of individuals to be sampled "
            "each time data is collected.")
            assert type(sampling_params.n) is int, ("The "
            "'n' data-sampling parameter must be an integer.")
        #TODO: Add more assert statements here to check that only the right
        #combinations of parameters can be provided

        #get the number of individuals to sample, if applicable
        self.n = None
        if sampling_params.scheme != 'all':
            self.n = sampling_params.n

        #calculate the transect points, if 'transect' is the chosen sampling
        #scheme, or just get the points if 'point' is the scheme
        self.pts = None
        if sampling_params.scheme == 'point':
            self.pts = sampling_params.points
        if sampling_params.scheme == 'transect':
            endpts = sampling_params.transect_endpoints
            n_transect_pts = sampling_params.n_transect_points
            pts = _get_transect_points(endpoints = endpts, n = n_transect_pts)
            self.pts = pts

        #create the point buffers, if either 'transect' or 'point' is the 
        #chosen sampling scheme
        self.pts_area = None
        if sampling_params.scheme in ['transect', 'point']:
            self.pts_area = _make_point_buffers(self.pts,
                                        sampling_params.radius)

        #get the 'include_landscape' param (defaults False)
        self.include_landscape = False
        if ('include_landscape' in sampling_params.keys() and
            type(sampling_params.include_landscape) is bool):
            self.include_landscape = sampling_params.include_landscape

        #get the 'include_fixed_sites' param (defaults False)
        self.include_fixed_sites = False
        if ('include_fixed_sites' in sampling_params.keys() and
            type(sampling_params.include_landscape) is bool):
            self.include_fixed_sites = sampling_params.include_fixed_sites

        #set the when attribute to the 'when' parameter
        self.when = sampling_params.when
        #check type- and value-validity of self.when, and update its value
        #as necessary
        assert type(self.when in (list, float, int, type(None)))
        #if it's a list, make sure no values are greater than final timestep
        if type(self.when) is list:
            assert np.array([n < self.T for n in self.when]).all(), (' Values '
                'provided for sampling times must be less than total '
                'model run-time.')
        #if it's a float, int, or None
        elif type(self.when) in (float, int, type(None)):
            #check value is less than or equal to last timestep (or None)
            assert self.when is None or self.when < self.T, ('Values '
            'provided for sampling times must be less than total '
            'model run-time.')
            #make it a list containing nothing (last timestep will be added),
            #if 0 or None
            if self.when in (0, None):
                self.when = []
            #make it a stepwise timestep list, if integer other than 0
            else:
                self.when = [*range(0, self.T, int(self.when))]
        #add the final timestep, if not already in self.when
        if (len(self.when) == 0 or self.when[-1] != self.T-1):
            self.when.append(self.T-1)
        #now turn the when attribute into an iterator
        self.when = iter(self.when)
        #and set next_t
        self._set_next_t()

        #grab the genetic data formats as a separate attribute
        self.gen_formats = format_params.gen_format
        #change the gen_formats attribute to a list if it came in as a string
        if type(self.gen_formats) == str:
            self.gen_formats = [self.gen_formats]
        #also grab the geographic data formats as a separate attribute
        self.geo_formats = [format_params.geo_vect_format]
        #and grab the raster format, if a raster is required
        #NOTE: added to a separate attribute because this is written per
        #timestep, not per species within timestep
        self.rast_format = None
        if (sampling_params.include_landscape
            and 'geo_rast_format' in format_params.keys()):
            self.rast_format = format_params.geo_rast_format

    #method to set self.next_t
    def _set_next_t(self):
        try:
            self.next_t = next(self.when)
        #if there are no further timestep values left then this is the last
        #timestep, so set self.next_t to None
        except StopIteration:
            assert self.next_t == self.T -1, ("Model._set_next_t() threw a "
            "StopIteration error, but the current value of Model.next_t is "
            "not the final timestep (instead, it is %i).\n\n") % self.next_t
            self.next_t = None

    #method to create filenames for genetic and geographic datafiles
    def _make_filenames(self, iteration, spp_name):
        filenames = []
        for att_name in ['gen_formats', 'geo_formats']:
            filenames.append(['mod-%s_it-%i_t-%i_spp-%s.%s' % (self.model_name,
              iteration, self.next_t, spp_name, self.file_extension_dict[fmt])
                        for fmt in getattr(self, att_name)])
        return(filenames)

    #a method to be called each timestep, which will collect needed
    #data and then write the data (if write_intermittent == True) if it's
    #the right timestep
    #TODO: CONSIDER NOMENCLATURE CHANGE HERE AND IN CLASS NAME!
    def _write_data(self, community, land, iteration):

        #if this timestep is scheduled for sampling
        if community.t == self.next_t:

            #for each species
            for spp in community.values():

                #TODO: Probably get rid of this conditional; should be
                #unnecessary, and is not a direct check of the in-sync
                #assumption at any rate
                #double-check that each species' timestep is in sync with
                #comm.t and is scheduled for sampling 
                if spp.t == self.next_t:

                    #get the data directory name for this timestep
                    dirname = os.path.join(os.getcwd(),
                          'GNX_mod-%s' % self.model_name,
                                    'it-%i' % iteration)

                    #get the subdirectory for this species
                    subdirname = os.path.join(dirname, 'spp-%s' % spp.name)

                    #and create (and its parent data directory, if needed)
                    os.makedirs(subdirname, exist_ok = True)

                    #get filenames
                    gen_files, geo_files = self._make_filenames(
                                    iteration = iteration, spp_name = spp.name)

                    #sample individuals according to the scheme defined 
                    sample = self._get_sample(spp)

                    #write files, if sample length > 0 
                    #(NOTE: otherwise, an empty file with "ZERO_SAMPLE" in the 
                    #name will be written, below)
                    if len(sample) > 0:

                        #save genetic data, if the spp has a
                        #genomic architeecture
                        if spp.gen_arch is not None:
                            #for each genetic data format to be written
                            for n, data_format in enumerate(self.gen_formats):

                                #format the data accordingly
                                data = self._format_gen_data(
                                    data_format = data_format, sample = sample,
                                                                    spp = spp)

                                #then write it to disk
                                gen_filepath = os.path.join(subdirname,
                                                            gen_files[n])
                                self._write_gendata(filepath = gen_filepath,
                                                            gen_data = data)

                        #also write the geodata for this spp
                        for n, data_format in enumerate(self.geo_formats):
                            #write the geodata to disk
                            geo_filepath = os.path.join(subdirname,
                                                            geo_files[n])
                            self._write_geodata(filepath = geo_filepath,
                                               data_format = data_format,
                                               sample = sample)

                    #if sample was empty, write a placeholder file with name
                    #"<base_filename>_ZERO_SAMPLE"
                    else:
                        filenames = [gen_files, geo_files]
                        filenames = [*chain.from_iterable(filenames)]
                        filename = os.path.splitext(filenames[0])[0]
                        filename = filename + '_ZERO_SAMPLE'
                        filepath = os.path.join(subdirname, filename)
                        self._write_gendata(filepath, '')

            #write the raster, if necessary
            if self.rast_format is not None:
                #for each Layer
                for lyr in land.values():
                    #get the raster filename
                    filename = 'mod-%s_it-%i_t-%i_lyr-%s.%s' % (
                        self.model_name, iteration, self.next_t, lyr.name,
                                    self.file_extension_dict[self.rast_format])
                    filepath = os.path.join(dirname, filename)
                    #and write it to disk
                    lyr._write_raster(filepath, self.rast_format)

            #update self.next_t to the next timestep to be sampled
            self._set_next_t()


    def _get_random_sample(self, individuals):
        if len(individuals) > self.n:
            sample = r.choice(individuals, size = self.n, replace = False)
        else:
            sample = individuals
        return(sample)


    def _get_point_sample(self, spp):
        #TODO: check if this should be sped up any more
        sample = [i for i,v in spp.items() if self.pts_area.contains(
                                                        Point(v.x, v.y))]
        if len(sample) > self.n:
            sample = self._get_random_sample(individuals = sample)
        return(sample)


    def _get_sample(self, spp):
        #get a set of indices for the individuals in the sample
        sample = set()
        #take the whole species, if scheme is 'all'
        if self.scheme == 'all':
            sample.update([*spp])
        #or take a random sample, if scheme is 'random'
        elif self.scheme == 'random':
            inds = self._get_random_sample([*spp])
            sample.update(inds)
        #or take individuals within a given radius of self.buffs (which could
        #have come from a set of input points or from a calculated set of
        #transect points), if scheme is 'point' or 'transect'
        elif self.scheme in ['point', 'transect']:
            inds = self._get_point_sample(spp)
            sample.update(inds)
        #sort sample, then convert to a dict of individuals
        sample = sorted([*sample])
        sample = {i:spp[i] for i in sample}
        return(sample)


    def _format_gen_data(self, data_format, sample, spp):
        '''<data_format> can be:
                            'fasta'
                            'vcf'
        '''
        genotypes = spp._get_genotypes(individs=[*sample], as_dict=True)
        if data_format == 'fasta':
            formatted_data = _format_fasta(sample, genotypes)
        elif data_format == 'vcf':
            formatted_data = _format_vcf(sample, genotypes, spp.gen_arch,
                    include_fixed_sites = self.include_fixed_sites)
        return(formatted_data)


    def _write_gendata(self, filepath, gen_data):
        _write_file(filepath, gen_data)


    def _write_geodata(self, filepath, data_format, sample):
        write_fn = self.write_geodata_fn_dict[data_format]
        write_fn(filepath = filepath, individuals = sample)


######################################
# -----------------------------------#
# FUNCTIONS -------------------------#
# -----------------------------------#
######################################

#a function to get a set of n evenly spaced points between endpoints
def _get_transect_points(endpoints, n):
    x_pts = np.linspace(endpoints[0][0] , endpoints[1][0], n)
    y_pts = np.linspace(endpoints[0][1] , endpoints[1][1], n)
    return(list(zip(x_pts, y_pts)))


#a function to make shapely geometry buffers around a set of points
def _make_point_buffers(points, radius):
    pts = [Point(*p) for p in points]
    buffs = [p.buffer(radius) for p in pts]
    cu = cascaded_union(buffs)
    return(cu)


# a function that returns an ad hoc sample of size n
# from the current population of the given spp,
# in the format index-keyed dict format needed by the data-writing functions
# (to be used by the convenience methods of the Model object)
def _get_adhoc_sample(spp, n):
    #get a set of indices for the individuals in the sample
    sample = set()
    #take the whole species, if scheme is 'all'
    if n is None:
        sample.update([*spp])
    #or take a random sample, if scheme is 'random'
    else:
        if len(spp) > n:
            inds = r.choice([*spp], size = n, replace = False)
        else:
            inds = [*spp]
        sample.update(inds)
    #sort sample, then convert to a dict of individuals
    sample = sorted([*sample])
    sample = {i:spp[i] for i in sample}
    return(sample)


def _format_fasta(sample, genotypes):
    '''
    FASTA FORMAT:

    >idx:haploid_num|x_location|y_location|phenotype0;phenotype1;...;
                    phenotypeN|env_var0;env_var1;...;env_varN|age|sex
    001110101010101010010101011101010110.....01011110

    '''
    # ensure that sample and genotypes have identical lenghts and orders
    assert len(sample) == len(genotypes), ("'sample', and 'genotypes' have"
                                           "different lenghts!")
    assert np.all([*sample] == [*genotypes]), ("'sample' and 'genotypes' do not"
                                               " have identical orders!")

    row1 = '>%s:HAP;%s;%s;%s;%s;%s;%s\n'
    file_text = ''

    for individ, genotype in zip(sample.values(), genotypes.values()):
        for hap in range(2):
            individ_row1 = re.sub('HAP', str(hap), row1)
            replace = tuple(map(lambda att: re.sub(',', '|', re.sub('[\[\] ]',
                '', str(getattr(individ, att)))), ['idx', 'x', 'y', 'age',
                                            'sex', 'z', 'e']))
            individ_row1 = individ_row1 % replace
            individ_row2 = ''.join([str(
                            base) for base in genotype[:,hap]]) + '\n'

            file_text = file_text + individ_row1 + individ_row2

    return(file_text)


def _format_vcf(sample, genotypes, gen_arch, include_fixed_sites=False):

    # ensure that sample and genotypes have identical lenghts and orders
    assert len(sample) == len(genotypes), ("'sample', and 'genotypes' have"
                                           "different lenghts!")
    assert np.all([*sample] == [*genotypes]), ("'sample' and 'genotypes' do not"
                                               " have identical orders!")

    #create a template header
        #NOTE: has 1 string slot for a date

        #TODO: DECIDE ON NECESSARY INFO AND FORMAT CONTENTS,
        #THEN ADD METADATA ROWS HERE

    header = '''##fileformat=VCFv4.2
##fileDate=%s
##source=Geonomics
'''

    #template column-header row
    col_header_row = ('#CHROM\tPOS\tID\tREF\tALT\tQUAL'
                      '\tFILTER\tINFO\tFORMAT\t%s\n')
        #NOTE: this has 1 string slot for a tab-separated
        #list of all individ ids

    #template data row
    #TODO: UPDATE/CHANGE THE INFO AND FORMAT PORTIONS OF THIS TEMPLATE,
    #AFTER I DECIDE ON THEIR CONTENTS (above)
    data_row = ('%i\t%i\t.\tA\tT\t1000\tPASS\t%s\tGT\t%s\n')
        #NOTE: this has 2 integer slots, then 1 string slot for:
            #- chrom number (NOTE: unpythonically, starts from 1)
            #- locus number (NOTE: reported cumulative from locus 0,
               #not from start of each chrom)
            #- a tab-separated list of individs' genotypes at this locus

    #create a col_header_row for this data
    inds = [*sample.keys()]
    ind_cols = '\t'.join([str(i) for i in inds])
    cols = col_header_row % (ind_cols)

    #get a list of the chromosome numbers
    #chroms = np.cumsum(gen_arch.l_c)

    #and get all individuals' genomic data in a 'samplome' object (a 3-d array)
    samplome = np.array([genotypes[i] for i in inds])

    #get all segregating sites
    max_val = 2 * len(sample)
    segs = np.where(samplome.sum(axis = 2).sum(axis = 0) > 0)[0]
    segs2 = np.where(samplome.sum(axis = 2).sum(axis = 0) < max_val)[0]
    segs = sorted(list(set(segs).intersection(set(segs2))))

    #if not_include_fixed_sites, include only the segregating sites in the VCF 
    if not include_fixed_sites:
        loci = segs
    #or else get all loci
    else:
        loci = range(gen_arch.L)

    #and get the sites' chrom nums
    #chroms = [list((locus - chroms) < 0).index(True) for locus in loci]

    #build all the VCF data rows
    rows = ''
    for n, locus in enumerate(loci):
        gts = samplome[:,locus,:]
        gts = '\t'.join(['%i|%i' % (gts[i,0],
                gts[i,1]) for i in range(np.shape(gts)[0])])
        #get an indicator for whether the site is segregating or fixed
        seg = locus in segs
        seg_dict = {True: 'SEG', False: 'FIX'}

        rows = rows + data_row % (0, locus, seg_dict[seg], gts)

    #get the date
    now = datetime.datetime.now()
    month = str(now.month).zfill(2)
    day = str(now.day).zfill(2)
    date = '%d%s%s' % (now.year, month, day)

    #paste all the VCF content together
    out_vcf = ''.join([header % date, cols, rows])

    #return it
    return(out_vcf)


def _tskit_table_to_pandas(table):
    """
    Takes a raw tskit.TableCollection table dict,
    as returned from tskit.TableCollection.<TABLENAME>.asdict(),
    and converts it into a pandas.DataFrame, for write-out
    """
    # get the max column length
    max_col_len = max([len(col) for col in table.values()])
    # fill in columns with np.nan
    filled_table = {}
    for k, col in table.items():
        if len(col) > 0:
            filled_table[k] = np.concatenate((col,
                                    np.nan * np.ones(max_col_len - len(col)))) 
        else:
            filled_table[k] = np.ones(max_col_len) * np.nan
    # convert to pandas.DataFrame
    pd_table = pd.DataFrame.from_dict(filled_table)
    return pd_table




#!/usr/bin/python
# IBD_IBE_test.py

# flake8: noqa


########
# set-up
########

# geonomics imports
from geonomics.utils.viz import _check_display

# other imports
from copy import deepcopy
# from itertools import chain
import numpy as np
from sklearn.decomposition import PCA
import os
import matplotlib as mpl
_check_display()
from matplotlib import animation
import matplotlib.pyplot as plt
from matplotlib import gridspec
from matplotlib.ticker import FormatStrFormatter
from mpl_toolkits.mplot3d import Axes3D
from copy import copy, deepcopy
import time
import statsmodels.api as sm




###############
# function defs
###############

def map_genetic_PCA(mod, mark_size):
    """run genetic PCA and mapping individuals colored in RGB
    space by the first 3 PCs"""
    species = mod.comm[0]
    land = mod.land
    # get array of resulting genomic data (i.e. 'speciome'),
    # genotypes meaned by individual
    speciome = np.mean(np.stack([i.g for i in species.values()]), axis=2)
    # run PCA on speciome
    pca = PCA(n_components=3)
    PCs = pca.fit_transform(speciome)
    # normalize the PC results
    norm_PCs = (PCs - np.min(PCs,
                             axis=0)) / (np.max(PCs,
                                                axis=0) - np.min(PCs,
                                                                 axis=0))
    # use first 3 PCs to get normalized values for R, G, & B colors
    PC_colors = norm_PCs * 255
    # scatter all individuals on top of landscape, colored by the
    # RBG colors developed from the first 3 geonmic PCs
    xs = mod.comm[0]._get_x()
    ys = mod.comm[0]._get_y()
    # get environmental raster, with barrier masked out
    masked_env = deepcopy(mod.land[0].rast)
    masked_env[mod.land[1].rast == 0] = np.nan
    # create light colormap for plotting landscape
    # bot = plt.cm.get_cmap('Blues', 256)(np.linspace(0.4, 0.45, 2))[0]
    # top = plt.cm.get_cmap('Reds', 256)(np.linspace(0.4, 0.45, 2))[0]
    cmap = plt.cm.coolwarm
    copy(cmap).set_bad(color='#8C8C8C')
    # plot landscape
    # plt.imshow(masked_env, cmap=cmap, alpha=0.8)
    mod.plot(0, 0, color=PC_colors/255.0, mask_rast=masked_env,
             size=mark_size, edge_color='black')
    #plt.pcolormesh(land._x_cell_bds, land._y_cell_bds, masked_env, cmap=cmap)
    # scatter plot of individuals, colored by composite PC score
    #plt.scatter(xs, ys, c=PC_colors/255.0, s=mark_size, edgecolors='black')
    # fix x and y limits
    #[f([dim - 0.5 for dim in (0, mod.land.dim[0])]) for f in (plt.xlim,
    #                                                          plt.ylim)]
    # get rid of x and y ticks
    #[f([]) for f in (plt.xticks, plt.yticks)]


def plot_genetic_PCA(mod):
    """run a genetic PCA and plot individuals on first 2 PCs, colored by which
    side of the barrier they are on"""
    from copy import deepcopy
    from sklearn.decomposition import PCA
    figsize = 6
    species = mod.comm[0]
    # get array of resulting genomic data (i.e.
    # 'speciome'), genotypes meaned by individual
    speciome = np.mean(np.stack([i.g for i in species.values()]), axis=2)
    # run PCA on speciome
    pca = PCA(n_components=2)
    PCs = pca.fit_transform(speciome)
    # normalize the PC results
    norm_PCs = (PCs - np.min(PCs,
                    axis=0)) / (np.max(PCs, axis=0) - np.min(PCs, axis=0))
    # assign a value to each species, 0 or 1, indicating whether
    # they're located on the left or right vertical half of the landscape
    ind_colors = [0 if (
                ind.x < mod.land.dim[0]/2) else 1 for ind in species.values()]
    # plot individuals on PCs 1 and 2, colored by their landscape half
    fig = plt.figure(figsize=(figsize, figsize), dpi= 80,
                     facecolor='w', edgecolor='k')
    ax1 = fig.add_subplot(111)
    PC_dict = dict(zip(range(len(ind_colors)), ind_colors))
    left = [k for k,v in PC_dict.items() if v == 0]
    right = [k for k,v in PC_dict.items() if v == 1]
    patches = []
    patches.append(ax1.scatter(norm_PCs[left, 0], norm_PCs[left, 1],
                               color = '#00ffff'))
    patches.append(ax1.scatter(norm_PCs[right, 0], norm_PCs[right, 1],
                               color = '#ff00ff'))
    ax1.set_xlabel('genetic PC 1')
    ax1.set_ylabel('genetic PC 2')
    ax1.legend(patches, ['left of barrier', 'right of barrier'])


# calculate euclidean distance from two n-length vectors
def calc_euc(x, y):
    euc = np.sqrt(sum([(n-m)**2 for n, m in zip(x, y)]))
    return euc


# calculate lower-triangular of PCA-based Euclidean genetic distances between
# all individuals, using a 'speciome' (2d array of all individs' genomes)
def calc_dists(species, dist_type='gen', env_lyrs=None, return_flat=True,
               allele_freq_diff=False, biallelic=False):
    # calculate genetic distance as the euclidean distance between individuals
    # in genetic PC space
    if dist_type == 'gen':
        speciome = species._get_genotypes(biallelic=biallelic)
        if not allele_freq_diff:
            pca = PCA()
            vals = pca.fit_transform(speciome)
        else:
            vals = speciome
    # calculate geographic distance as the linear euclidean distance between
    # individuals
    elif dist_type == 'geo':
        vals = np.stack([np.array((i.x, i.y)) for i in species.values()])
    # calculate environmental distance as the euclidean distance between
    # individuals' environments, for all environmental layers specified by
    # the 'env_lyrs' argument
    elif dist_type == 'env':
        vals = np.stack([np.array(i.e)[env_lyrs] for i in species.values()])
    #calculate the phenotypic distances
    elif dist_type == 'phn':
        vals = np.stack([np.array(i.z)[env_lyrs] for i in species.values()])
    # print(vals)
    # print(vals.shape)
    n_ind = vals.shape[0]
    dist_mat = np.ones([n_ind] * 2) * np.nan
    # dist_vals = [[calc_euc(i, j) for j in vals[n:,
    #                                        :]] for n, i in enumerate(vals)]
    for i in range(n_ind):
        for j in range(0, i+1):
            if dist_type == 'gen' and allele_freq_diff:
                dist_mat[i, j] = np.sum(np.abs(
                                        vals[:, i] - vals[:, j])) / vals.shape[1]
            else:
                dist_mat[i, j] = calc_euc(vals[i, :], vals[j, :])
    # check that all diagonal values are 0
    assert np.all(np.diag(dist_mat) == 0), "Not all diagonal values are 0!"
    if dist_type == 'gen' and allele_freq_diff:
        assert (np.all(0 <= dist_mat[np.invert(np.isnan(dist_mat))])
                and np.all(dist_mat[np.invert(np.isnan(dist_mat))] <= 1)), (
            "You are calculating genetic distances using allele-frequency "
            "differences, which should be normalized to 0 <= diff <= 1, but "
            "you have non-NaN values outside that range!")

    if return_flat:
        # flatten the lower triangle, if return 1-d of values
        dists = dist_mat[np.tril_indices(dist_mat.shape[0], -1)]
        assert dists.size == (n_ind**2 - n_ind)/2, ("Length not equal "
                                                    "to n(n-1)/2!")
    else:
        # make it a symmetric dist matrix, if returning the matrix
        dist_mat[np.isnan(dist_mat)] = 0
        dists = dist_mat + dist_mat.T
        assert dists.size == n_ind**2, "Size not equal to n*n!"
    # dist_vals = [item[1:] for item in dist_vals]
    # dists = [*chain.from_iterable(dist_vals)]
    # assert that the length is correct
    return dists


# define a function to calculate the mean difference between phenotype
# and environment for a species
def calc_mean_z_e_diff(spp, trait_num=0):
    zs = spp._get_z().ravel()
    es = spp._get_e(lyr_num=spp.gen_arch.traits[trait_num].lyr_num)
    mean_diff = np.mean(np.abs(zs - es))
    return mean_diff


# define a function to calculate individuals' mean fitness
def calc_mean_fitness(spp):
    return np.mean(spp._calc_fitness())


def track_horiz_crossing(mod, zone_edges, tracker, count):
    """track all horizontal crossing of some vertical zone on the landscape,
    delineated by x-axis bounds in zone_edges argument, so that the model can
    be run with and without the barrier mask, to quantify the extent to which
    the barrier increases landscape resistance
    """
    for idx, ind in mod.comm[0].items():
        # individ is assigned value 0 if on left side of the crossing zone...
        if ind.x <= zone_edges[0]:
            curr_side = 0
        #...or 1 if on right side...
        elif ind.x >= zone_edges[1]:
            curr_side = 1
        # or else nan if they are inside the zone
        else:
            curr_side = np.nan
        # if individ is in the tracker then they were alive last time step,
        # so grab and compare their previous value to this time step's value
        if idx in tracker:
            diff = curr_side - tracker[idx]
            # ignore nans, becuase they are currently inside the zone, so
            # whether or not they are crossing cannot be decided
            if np.isnan(diff):
                pass
            # increment the crossing counter, if a crossing has occurred
            elif diff != 0:
                count +=1
        # now update the individ's value in the tracker (unless they have a
        # nan, because they're currently inside the zone, so maintain the
        # value from their last time outside the zone until they exit
        # the zone and it can be determined whether or not they crossed
        if not np.isnan(curr_side):
            tracker[idx] = curr_side
    return count


# function for setting a precise number of axis ticks, rounded to a fixed
# number of digits
def fix_yax_n_ticks_digits(ax, vals, n_ticks, n_digits):
    minval = min(vals)
    maxval = max(vals)
    tick_vals = np.linspace(minval, maxval, n_ticks)
    tick_labs = [('%%0.%if' % n_digits) % v for v in tick_vals]
    ax.set_yticks(tick_vals)
    ax.set_yticklabels(tick_labs)


# set some plotting params
img_dir = ('/home/drew/Desktop/stuff/berk/research/projects/sim/methods_paper/'
           'img/final/')
ax_fontdict = {'fontsize': 65,
               'name': 'Bitstream Vera Sans'}
ze_ax_fontdict = {'fontsize': 35,
                  'name': 'Bitstream Vera Sans'}
threed_ax_fontdict = {'fontsize': 35,
                  'name': 'Bitstream Vera Sans'}
ttl_fontdict = {'fontsize': 65,
                'name': 'Bitstream Vera Sans'}
mark_size = 275


def _make_params():

    # IBD_IBE_params.py

    # This is a parameters file generated by Geonomics
    # (by the gnx.make_parameters_file() function).


                       ##  ::::::          :::    :: ::::::::::##
                 ##:::::    ::::   :::      ::    :: :: ::::::::::: :##
              ##::::::::     ::            ::   ::::::::::::::::::::::::##
            ##:::::::::                      :::::::::: :::::: ::::::::  :##
          ## : ::::  ::                    ::::  : ::    :::::::: : ::  :   ##
         ##GGGGG  EEEEE OOOOO   NN   NN   OOOOO   MM   MM IIIIII  CCCCC SSSSS##
        ##GG     EE    OO   OO  NNN  NN  OO   OO  MM   MM   II   CC     SS    ##
        ##GG     EE   OO     OO NN N NN OO     OO MMM MMM   II   CC     SSSSSS##
        ##GG GGG EEEE OO     OO NN  NNN OO     OO MM M MM   II   CC         SS##
        ##GG   G EE    OO   OO  NN   NN  OO   OO  MM   MM   II   CC        SSS##
         ##GGGGG  EEEEE OOOOO   NN   NN   OOOOO   MM   MM IIIIII  CCCCC SSSSS##
          ##    :::::::::               :::::::::: ::              ::  :   :##
            ##:   :::::                    :::::: :::             ::::::: ##
              ##   :::                      :::::  ::              :::::##
                 ## ::                      ::::                     ##
                       ##                                      ##
                          ## :: ::    :::            ##


    env_left_half = np.hstack([np.atleast_2d(np.linspace(
                    0, 1, 40) + np.random.normal(
                    0, 0.05, 40)).T for _ in range(20)])
    env_right_half = np.hstack([np.atleast_2d(np.linspace(
                    0, 1, 40) + np.random.normal(
                    0, 0.05, 40)).T for _ in range(20)])
    env_right_half = np.flipud(env_right_half)
    env = np.hstack((env_left_half, env_right_half))
    env = np.clip(env, 0, 1)
    barrier = np.ones((40, 40))
    barrier[:, 18:22] = 0


    params = {
    ###############################################################################

    ###################
    #### LANDSCAPE ####
    ###################
        'landscape': {

        ##############
        #### main ####
        ##############
            'main': {
                #y,x (a.k.a. i,j) dimensions of the Landscape
                'dim':                      (40,40),
                #resolution of the Landscape
                'res':                      (1,1),
                #upper-left corner of the Landscape
                'ulc':                      (0,0),
                #projection of the Landscape
                'prj':                      None,
                }, # <END> 'main'

        ################
        #### layers ####
        ################
            'layers': {

                #layer name (LAYER NAMES MUST BE UNIQUE!)
                'env': {

            #######################################
            #### layer num. 0: init parameters ####
            #######################################

                    #initiating parameters for this layer
                    'init': {
                        #parameters for a 'defined'-type Layer
                        'defined': {
                            'rast': env,
                            'pts': None,
                            'vals': None,
                            'interp_method': None
                            }, # <END> 'defined'

                        }, # <END> 'init'

                    }, # <END> layer num. 0

                #layer name (LAYER NAMES MUST BE UNIQUE!)
                'barrier': {

            #######################################
            #### layer num. 1: init parameters ####
            #######################################

                    #initiating parameters for this layer
                    'init': {
                        #parameters for a 'defined'-type Layer
                        'defined': {
                            'rast': barrier,
                            'pts': None,
                            'vals': None,
                            'interp_method': None
                            }, # <END> 'defined'

                        }, # <END> 'init'

                    }, # <END> layer num. 0



        #### NOTE: Individual Layers' sections can be copy-and-pasted (and
        #### assigned distinct keys and names), to create additional Layers.


                } # <END> 'layers'

            }, # <END> 'landscape'


    ###############################################################################

    ###################
    #### COMMUNITY ####
    ###################
        'comm': {

            'species': {

                #species name (SPECIES NAMES MUST BE UNIQUE!)
                'spp_0': {

                #####################################
                #### spp num. 0: init parameters ####
                #####################################

                    'init': {
                        #starting number of individs
                        'N':                1000,
                        #carrying-capacity Layer name
                        'K_layer':          'barrier',
                        #multiplicative factor for carrying-capacity layer
                        'K_factor':         1.5
                        }, # <END> 'init'

                #######################################
                #### spp num. 0: mating parameters ####
                #######################################

                    'mating'    : {
                        #age(s) at sexual maturity (if tuple, female first)
                        'repro_age':                0,
                        #whether to assign sexes
                        'sex':                      False,
                        #ratio of males to females
                        'sex_ratio':                1/1,
                        #intrinsic growth rate
                        'R':                        0.5,
                        #intrinsic birth rate (MUST BE 0<=b<=1)
                        'b':                        0.5,
                        #expectation of distr of n offspring per mating pair
                        'n_births_distr_lambda':    1,
                        #whether n births should be fixed at n_births_dist_lambda
                        'n_births_fixed':           True,
                        #radius of mate-search area
                        'mating_radius':            2,
                        'inverse_dist_mating':      False,
                        'choose_nearest_mate':      False,
                        }, # <END> 'mating'

                ##########################################
                #### spp num. 0: mortality parameters ####
                ##########################################

                    'mortality'     : {
                        #maximum age
                        'max_age':                      None,
                        #min P(death) (MUST BE 0<=d_min<=1)
                        'd_min':                        0,
                        #max P(death) (MUST BE 0<=d_max<=1)
                        'd_max':                        1,
                        #width of window used to estimate local pop density
                        'density_grid_window_width':    None,
                        }, # <END> 'mortality'

                #########################################
                #### spp num. 0: movement parameters ####
                #########################################

                    'movement': {
                        #whether or not the species is mobile
                        'move':                     True,
                        #mode of distr of movement direction
                        'direction_distr_mu':       1,
                        #concentration of distr of movement direction
                        'direction_distr_kappa':    0,
                        #mean of distr of movement distance
                        'movement_distance_distr_param1':        0.5,
                        #variance of distr of movement distance
                        'movement_distance_distr_param2':     0.5,
                        'movement_distance_distr':            'wald',
                        #first param of distr of dispersal distance
                        'dispersal_distance_distr_param1':       0.5,
                        #second param of distr of dispersal distance
                        'dispersal_distance_distr_param2':    0.5,
                        'dispersal_distance_distr':  'wald',
                        'move_surf'     : {
                            #move-surf Layer name
                            'layer': 'barrier',
                            #whether to use mixture distrs
                            'mixture': True,
                            #concentration of distrs
                            'vm_distr_kappa': 12,
                            #length of approximation vectors for distrs
                            'approx_len': 5000,
                            },  # <END> 'move_surf'

                        },    # <END> 'movement'


                #####################################################
                #### spp num. 0: genomic architecture parameters ####
                #####################################################

                    'gen_arch': {
                        #file defining custom genomic arch
                        'gen_arch_file':            None,
                        #num of loci
                        'L':                        100,
                        #num of chromosomes
                        'l_c':                      [100],
                        #whether starting allele frequencies should be fixed at 0.5
                        'start_p_fixed':            0.5,
                        #genome-wide per-base neutral mut rate (0 to disable)
                        'mu_neut':                  0,
                        #genome-wide per-base deleterious mut rate (0 to disable)
                        'mu_delet':                 0,
                        #shape of distr of deleterious effect sizes
                        'delet_alpha_distr_shape':  0.2,
                        #scale of distr of deleterious effect sizes
                        'delet_alpha_distr_scale':  0.2,
                        #alpha of distr of recomb rates
                        'r_distr_alpha':            None,
                        #beta of distr of recomb rates
                        'r_distr_beta':             None,
                        #whether loci should be dominant (for allele '1')
                        'dom':                      False,
                        #whether to allow pleiotropy
                        'pleiotropy':               False,
                        #custom fn for drawing recomb rates
                        'recomb_rate_custom_fn':    None,
                        #number of recomb paths to hold in memory
                        'n_recomb_paths_mem':       int(1e4),
                        #total number of recomb paths to simulate
                        'n_recomb_paths_tot':       int(1e5),
                        'n_recomb_sims':            10000,
                        'start_neut_zero':          False,
                        'allow_ad_hoc_recomb':      False,
                        'jitter_breakpoints':       False,
                        #whether to save mutation logs
                        'mut_log':                  False,

                        'traits': {

                            ###########################
                            ####trait 0 parameters ####
                            ###########################
                            #trait name (TRAIT NAMES MUST BE UNIQUE!)
                            'trait_0': {
                                #trait-selection Layer name
                                'layer':                'env',
                                #polygenic selection coefficient
                                'phi':                  0.05,
                                #number of loci underlying trait
                                'n_loci':               10,
                                #mutation rate at loci underlying trait
                                'mu':                   0,
                                #mean of distr of effect sizes
                                'alpha_distr_mu':      0.1,
                                #variance of distr of effect size
                                'alpha_distr_sigma':    0,
                                #max alpha value
                                'max_alpha_mag':        None,
                                #curvature of fitness function
                                'gamma':                1,
                                #whether the trait is universally advantageous
                                'univ_adv':             False
                                }, # <END> trait 0


        #### NOTE: Individual Traits' sections can be copy-and-pasted (and
        #### assigned distinct keys and names), to create additional Traits.


                            }, # <END> 'traits'

                        }, # <END> 'gen_arch'


                    }, # <END> spp num. 0



        #### NOTE: individual Species' sections can be copy-and-pasted (and
        #### assigned distinct keys and names), to create additional Species.


                }, # <END> 'species'

            }, # <END> 'comm'


    ###############################################################################

    ###############
    #### MODEL ####
    ###############
        'model': {
            #total Model runtime (in timesteps)
            'T':            10,
            #min burn-in runtime (in timesteps)
            'burn_T':       30,
            #seed number
            'num':          None,
            #time step interval for simplification of tskit tables
            'tskit_simp_interval':      100,

            ###############################
            #### iterations parameters ####
            ###############################
            'its': {
                #num iterations
                'n_its':            1,
                #whether to randomize Landscape each iteration
                'rand_landscape':   False,
                #whether to randomize Community each iteration
                'rand_comm':        False,
                #whether to burn in each iteration
                'repeat_burn':      False,
                }, # <END> 'iterations'



            } # <END> 'model'

        } # <END> params

    return params


def _run(params, save_figs=False, time_it=False,
         make_3d_plots=False, use_barrier=True,
         run_mantel=False):
    """
    use_barrier:
        sets flag to indicate whether or not to use the barrier
        (this will typically be used, but can be turned off to run the model
        without a barrier and thus determine, comparatively, how much the
        barrier increases landscape resistance)
    """

    ###############
    # set up figure
    ###############
    nrows = 13
    ncols = 4
    ratio = ncols/nrows
    vert_size = 11 # length of figure's vertical dimension
    #fig = plt.figure(figsize=(vert_size, vert_size*ratio))
    plt.subplots_adjust(left=0.05, bottom=0.07, right=0.98, top=0.96, wspace=0.07,
                        hspace=0.16)
    gs = gridspec.GridSpec(nrows, ncols)
                          #height_ratios=[1]*8 + [1.5]*6,
                          #width_ratios=[1,1,1,0.7,0.7,1,1,1])

    #fig bounds
    gen_b4_top = 0
    gen_b4_bot = 3
    gen_b4_L = 0
    gen_b4_R = 2

    gen_af_top = 0
    gen_af_bot = 3
    gen_af_L = 2
    gen_af_R = 4

    ze_fit_top = 6
    ze_fit_bot = 10
    ze_fit_L = 1
    ze_fit_R = 3

    phn_b4_top = 3
    phn_b4_bot = 6
    phn_b4_L = 0
    phn_b4_R = 2

    phn_af_top = 3
    phn_af_bot = 6
    phn_af_L = 2
    phn_af_R = 4

    n1_3d_top = 10
    n1_3d_bot = 13
    n1_3d_L = 0
    n1_3d_R = 2

    n3_3d_top = 10
    n3_3d_bot = 13
    n3_3d_L = 2
    n3_3d_R = 4

    # BEFORE-SIM AXES
    gen_b4_ax = plt.subplots(1,1)[1]
    #gen_b4_ax = fig.add_subplot(gs[gen_b4_top:gen_b4_bot,
    #                               gen_b4_L:gen_b4_R], aspect='equal')
    gen_b4_ax.set_ylabel('genotype', fontdict=ax_fontdict)
    gen_b4_ax.set_title('before simulation', fontdict=ttl_fontdict)
    phn_b4_ax = plt.subplots(1,1)[1]
    #phn_b4_ax = fig.add_subplot(gs[phn_b4_top:phn_b4_bot,
    #                               phn_b4_L:phn_b4_R], aspect='equal')
    phn_b4_ax.set_ylabel('phenotype', fontdict=ax_fontdict)

    # AFTER-SIM AXES
    gen_af_ax = plt.subplots(1,1)[1]
    #gen_af_ax = fig.add_subplot(gs[gen_af_top:gen_af_bot,
    #                               gen_af_L:gen_af_R], aspect='equal')
    gen_af_ax.set_title('after simulation', fontdict=ttl_fontdict)
    #gen_af_ax.set_ylabel('genotype', fontdict=ax_fontdict)
    phn_af_ax = plt.subplots(1,1)[1]
    #phn_af_ax = fig.add_subplot(gs[phn_af_top:phn_af_bot,
    #                               phn_af_L:phn_af_R], aspect='equal')
    #phn_af_ax.set_ylabel('phenotype', fontdict=ax_fontdict)

    #--------
    # 3D AXES
    #--------
    # num 1
    n1_3d_ax = plt.figure().add_subplot(111, projection='3d')
    #n1_3d_ax = fig.add_subplot(gs[n1_3d_top:n1_3d_bot,
                                  #n1_3d_L:n1_3d_R], projection='3d')
    n1_3d_ax.view_init(elev=3, azim=83)
    n1_3d_ax.set_xlabel('$\longleftarrow$ geo. dist.', labelpad=-13,
                        fontdict=threed_ax_fontdict)
    #n1_3d_ax.set_ylabel(' ' * 35 + '$\longleftarrow$ Env. Dist.', size=9,
    #                    labelpad=20)
    n1_3d_ax.zaxis.set_rotate_label(False)
    n1_3d_ax.set_zlabel('gen. dist. $\longrightarrow$', labelpad=-13,
                        rotation=90, fontdict=threed_ax_fontdict)
    n1_3d_ax.set_xticklabels([])
    n1_3d_ax.set_yticklabels([])
    n1_3d_ax.set_zticklabels([])
    n1_3d_ax.set_title("", pad=-260)
    # num 2
    ze_fit_ax_L = plt.subplots(1,1)[1]
    #ze_fit_ax_L = fig.add_subplot(gs[ze_fit_top:ze_fit_bot,
    #                                 ze_fit_L:ze_fit_R])
    ze_fit_ax_R = ze_fit_ax_L.twinx()
    #n2_3d_ax = fig.add_subplot(gs[2, 3:6], projection='3d')
    #n2_3d_ax.view_init(elev=25, azim=225)
    #n2_3d_ax.set_xlabel('Geo. Dist. $\longrightarrow$', size=9, labelpad=-13)
    #n2_3d_ax.set_ylabel('$\longleftarrow$ Env. Dist.', size=9, labelpad=-13)
    #n2_3d_ax.zaxis.set_rotate_label(False)
    #n2_3d_ax.set_zlabel('Gen. Dist. $\longrightarrow$', size=9, labelpad=-13,
    #                    rotation=90)
    #n2_3d_ax.set_xticklabels([])
    #n2_3d_ax.set_yticklabels([])
    #n2_3d_ax.set_zticklabels([])

    # num 3
    n3_3d_ax = plt.figure().add_subplot(111, projection='3d')
    #n3_3d_ax = fig.add_subplot(gs[n3_3d_top:n3_3d_bot,
    #                              n3_3d_L:n3_3d_R], projection='3d')
    n3_3d_ax.view_init(elev=3, azim=7)
    n3_3d_ax.set_zlabel('gen. dist. $\longrightarrow$', labelpad=-13,
                        rotation=90, fontdict=threed_ax_fontdict)
    n3_3d_ax.set_ylabel('env. dist. $\longrightarrow$', labelpad=-13,
                        fontdict=threed_ax_fontdict)
    n3_3d_ax.zaxis.set_rotate_label(False)
    #n3_3d_ax.set_zlabel('gen. dist. $\longrightarrow$', size=12, labelpad=-13,
    #                    rotation=90)
    n3_3d_ax.set_xticklabels([])
    n3_3d_ax.set_yticklabels([])
    n3_3d_ax.set_zticklabels([])



    #########################
    # set trackers and timers
    #########################

    # create objects for crossing of barrier, and of equal-width vertical areas
    # within each of the two sides for comparison
    cross_count = 0
    cross_tracker = {}

    # start timer
    if time_it:
        start = time.time()


    #####################
    # prep and make model
    #####################

    # set model name (since the params are'nt being read in from separate file)
    params.model['name'] = 'IBD_IBE_demo'

    # get barrier-zone edges, for tracking crossings
    barr_rast = params['landscape']['layers']['barrier']['init']['defined']['rast']
    zone_edges = np.where(barr_rast[0,:] == 0)[0]
    zone_edges = (zone_edges.min(), zone_edges.max() + 1)

    # then get rid of barrier, if not to be used for this run
    if not use_barrier:
        params['landscape']['layers']['barrier']['init']['defined']['rast'] = np.ones(
                                            params['landscape']['main']['dim'])

    from .. import make_model
    mod = make_model(params)




    ################################
    # run model, plotting as it goes
    ################################

    # define number of timesteps
    T = 1000

    # burn model in
    mod.walk(20000, 'burn')

    # plot genetic PCA before evolution begins
    plt.sca(gen_b4_ax)
    map_genetic_PCA(mod, mark_size)

    # plot phenotypes before evolution begins
    mask = np.ma.masked_where(mod.land[1].rast == 0, mod.land[1].rast)
    plt.sca(phn_b4_ax)
    mod.plot_phenotype(0, 0, mask_rast=mask, size=mark_size, cbar=False)
    [f((-0.5, 39.5)) for f in [phn_b4_ax.set_xlim, phn_b4_ax.set_ylim]]

    # create data structure to save z-e diff values
    if not time_it:
        mean_z_e_diffs = []
        mean_fits = []

    # run model for T timesteps
    for t in range(T):
        if not time_it:
            mean_z_e_diffs.append(calc_mean_z_e_diff(mod.comm[0]))
            mean_fits.append(calc_mean_fitness(mod.comm[0]))
            cross_count = track_horiz_crossing(mod, zone_edges, cross_tracker,
                                               cross_count)
        mod.walk(1)
    if not time_it:
        mean_z_e_diffs.append(calc_mean_z_e_diff(mod.comm[0]))
        mean_fits.append(calc_mean_fitness(mod.comm[0]))
        cross_count = track_horiz_crossing(mod, zone_edges, cross_tracker,
                                           cross_count)

    # finish timer
    if time_it:
        stop = time.time()
        tot_time = stop - start


    # plot genetic PCA after 1/4T timesteps
    plt.sca(gen_af_ax)
    map_genetic_PCA(mod, mark_size)

    # plot the individuals' phenotypes
    plt.sca(phn_af_ax)
    mod.plot_phenotype(0, 0, mask_rast=mask, size=mark_size, cbar=False)
    [f((-0.5, 39.5)) for f in [phn_af_ax.set_xlim, phn_af_ax.set_ylim]]



    #########################
    # create 3d IBD, IBE plot
    #########################

    spp_subset = deepcopy(mod.comm[0])
    rand_inds = spp_subset._get_random_individuals(250)
    all_inds = [*spp_subset]
    for ind in all_inds:
        if ind not in rand_inds:
            spp_subset.pop(ind)
    gen_dists = calc_dists(spp_subset, biallelic=False, allele_freq_diff=False)
    # output the genetic distance matrix to CSV, if Mantel test required
    if run_mantel:
        gen_dist_mat = calc_dists(spp_subset, biallelic=False,
                                  allele_freq_diff=False,return_flat=False)
        np.savetxt("gen.csv", gen_dist_mat, delimiter=',')
    scaled_gen_dists = gen_dists/gen_dists.max()
    assert (np.all(scaled_gen_dists >= 0)
            and np.all(scaled_gen_dists <= 1)), ('Scaled genetic dist is outside '
                                                 '0 and 1!')
    geo_dists = calc_dists(spp_subset, 'geo')
    env_dists = calc_dists(spp_subset, 'env', [0])
    pheno_dists = calc_dists(spp_subset, 'phn', [0])

    # output the geo and env distance matrices to CSV, if Mantel test required,
    # then run the Mantel test
    if run_mantel:
        geo_dist_mat = calc_dists(spp_subset, 'geo', return_flat=False)
        env_dist_mat = calc_dists(spp_subset, 'env', [0], return_flat=False)
        np.savetxt("geo.csv", geo_dist_mat, delimiter=',')
        np.savetxt("env.csv", env_dist_mat, delimiter=',')
        # get the path to the mantel script, then run it
        gnx_path = '/' + os.path.join(*os.path.split(__file__)[0].split('/')[:-1])
        mantel_script_dir = os.path.join(gnx_path, 'data/IBD_IBE_demo')
        mantel_script = os.path.join(mantel_script_dir,
                                     os.listdir(mantel_script_dir)[0])
        print('\n\n')
        os.system('Rscript %s' % mantel_script)
        print('\n\n')

    # run multiple linear regression of gen on geo and env dists
    mlr_est = sm.Logit(endog=np.array(scaled_gen_dists.T),
                       exog=np.vstack((geo_dists, env_dists)).T).fit()

    # create logistic regression (to use in wireframe predicted surface)
    x_preds = np.arange(0, 50.1, 0.1)
    y_preds = np.linspace(0, 1, len(x_preds))
    z_preds = mlr_est.predict(np.vstack((x_preds, y_preds)).T)
    p_val_lt = mlr_est.pvalues[0] < 0.001
    assert p_val_lt, 'p-value not less than 0.001!'
    print(mlr_est.summary2())

    # create 3d-plot data
    y_vals = np.arange(0, 1.01, 0.01)
    ys = np.hstack([list(y_vals) for _ in range(len(y_vals))])
    xs = np.hstack([[n] * len(y_vals) for n in np.linspace(0, 50, len(y_vals))])
    zs = mlr_est.predict(np.vstack((xs, ys)).T)
    xs = xs.reshape([len(y_vals)] * 2)
    ys = ys.reshape([len(y_vals)] * 2)
    zs = zs.reshape([len(y_vals)] * 2)
    col3d = pheno_dists / pheno_dists.max()


    # plot on 3d axes
    for ax in [n1_3d_ax, n3_3d_ax]:#, n2_3d_ax]:
       ax.scatter(geo_dists, env_dists, scaled_gen_dists,
                  alpha=0.7, edgecolor='black', c=col3d, cmap='plasma')

    ##########################
    # create plot of z-e diffs
    ##########################
    if not time_it:
        L_color = '#096075'
        R_color = '#bf2659'
        ze_fit_ax_L.set_xlabel('time (steps)', fontdict=ze_ax_fontdict)
        ze_fit_ax_L.set_ylabel('mean($|z-e|$)', fontdict=ze_ax_fontdict,
                               color=L_color)
        ze_fit_ax_L.tick_params(axis='y', labelcolor=L_color, labelrotation=45,
                               labelsize=30)
        ze_fit_ax_L.tick_params(axis='x', labelsize=30)
        ze_fit_ax_L.plot(range(len(mean_z_e_diffs)), mean_z_e_diffs, color=L_color)
        fix_yax_n_ticks_digits(ze_fit_ax_L, mean_z_e_diffs, 5, 2)
        ze_fit_ax_R.set_ylabel('mean fitness', color=R_color,
                               fontdict=ze_ax_fontdict)
        ze_fit_ax_R.tick_params(axis='y', labelcolor=R_color,
                                labelrotation=-45, labelsize=30)
        ze_fit_ax_R.plot(range(len(mean_fits)), mean_fits, color=R_color)
        fix_yax_n_ticks_digits(ze_fit_ax_R, mean_fits, 5, 3)
        #ax.set_xlabel('time')
        #ax.set_ylabel(('mean difference between individuals\' phenotypes and '
        #               'environmental values'))
        #z_e_fig.show()
        #if save_figs:
        #    fig.savefig('IBD_IBE_z-e_plot.png', format='png', dpi=1000)

        #fig.tight_layout()
        #fig.show()
        if save_figs:
                fig.savefig('IBD_IBE.png', format='png', dpi=1000)


    #########################
    # create 3d animated plot
    #########################
    if make_3d_plots:
        plt.rc('animation', html='html5')
        fig3d = plt.figure()
        ax3d = fig3d.add_subplot(111, projection='3d')
        ax3d.set_xlabel('geo', size=15)
        ax3d.set_ylabel('env', size=15)
        ax3d.set_zlabel('gen', size=15)
        # initialization function, which plots the background of each frame
        def init():
            ax3d.plot_wireframe(xs, ys, zs, color='lightgray', ccount=10, rcount=10,
                                linestyles='dashed', alpha=0.95, linewidths=0.5)
            scat = ax3d.scatter(geo_dists, env_dists, scaled_gen_dists,
                                 alpha=0.5, c=col3d, cmap='plasma')
            return fig3d,

        # animation function, which is called sequentially
        def animate(i):
            ax3d.view_init(elev=5., azim=i)
            return fig3d,

        # use the init and animate functions to create an animation object
        anim = animation.FuncAnimation(fig3d, animate, init_func=init, frames=359,
                                       interval=5, blit=True)
        # write to file
        if save_figs:
            try:
                anim.save('IBD_IBE_animation.gif', writer='imagemagick',
                          fps=60)
            except Exception as e:
                print(('\nCould not use Imagemagick to save the 3D plot\'s '
                       'GIF. The following error was thrown:\n\t%s') % e)

        fig3d.show()


    #####################################
    # print out the zone-crossing results
    #####################################
    cross_rate = cross_count / T
    print(("\nThe barrier-zone crossing rate was %0.6f individuals "
          "per time step.\n") % cross_rate)

    # print out time
    if time_it:
        print("\n\nModel ran in %0.2f seconds." % tot_time)

    return mod




#!/usr/bin/python
# _str_repr_.py

'''
Defines core functions to aid in formatting strings for STDOUT
'''


######################################
# -----------------------------------#
# FUNCTIONS -------------------------#
# -----------------------------------#
######################################

#function to get the spaces to insert between an item and its value for a line
#of output in a class' __repr__ string
def _get_str_spacing(unformatted_str):
    tot_spacing = 48
    spacing_len = tot_spacing - (1 + len(unformatted_str.split(':')[0]))
    spacing = ' ' * spacing_len
    return spacing




#!/usr/bin/python
# spatial.py

'''
Defines core spatial classes, with their associated methods and supporting
functions
'''

# geonomics imports

# other imports
import numpy as np
import numpy.random as r
from numpy import pi
from scipy.stats import vonmises as s_vonmises
from collections import Counter as C
from copy import deepcopy
from operator import itemgetter as ig
from scipy import interpolate
from scipy.spatial import cKDTree
from shapely import geometry as g

try:
    from nlmpy import nlmpy
    with_nlmpy = True
except ModuleNotFoundError:
    print(("NOTE: Module 'nlmpy' not found. To simulate on Neutral "
           "Landscape Model rasters, this must be installed. "
           "To install, try using `pip install nlmpy`."))
    with_nlmpy = False

# set default vonmises params
s_vonmises.a = -np.inf
s_vonmises.b = np.inf

# #####################################
# -----------------------------------#
# CLASSES ---------------------------#
# -----------------------------------#
# #####################################


class _DensityGrid:
    def __init__(self, dim, dim_om, window_width, gi, gj, cells, areas,
                 x_edge, y_edge):

        # same as land.dim
        self.dim = dim

        # resolution (i.e. cell-size); defaults to 1,1
        self.res = (1, 1)

        # order of magnitude of the largest dimension in self.dim
        # (used for zero-padding the cell strings)
        self.dim_om = dim_om

        # window width to be used for grid-cell windows within which
        # species will be counted
        self.window_width = window_width

        # self.x_edge/y_edge are True if this grid has cells centered
        # on the edge (i.e. 0 and the dim val) for this dimension, else False
        self.x_edge = x_edge
        self.y_edge = y_edge

        # meshgrid arrays of the i and j points associated with each window
        self.gi = gi
        self.gj = gj

        # create array of the i,j grid-cell centerpoints
        self.grid_coords = np.array(list(zip(self.gi.flatten(),
                                             self.gj.flatten())))

        # create attributes of the grid cells and their areas
        self.cells = cells
        self.areas = areas

        # save itemgetter as an attribute (returns the counted
        # number of instances of each cell from the collections.Counter() dict)
        self.grid_counter = ig(*self.cells)

    def _calc_density(self, x, y):
        # determine the x and y cells within which each individual's
        # x and y coordinates fall (the self.x_edge and self.y_edge
        # corrections determine the correct cells whether the grid
        # includes cells centered on the landscape edges,
        # i.e. self.x_edge or y_edge = True, or not)
        x_cells = (x - self.x_edge * (
            self.window_width) / 2.) // self.window_width + self.x_edge
        y_cells = (y - self.y_edge * (
            self.window_width) / 2.) // self.window_width + self.y_edge

        # get cell strings from indivudals' cells
        cells = _make_cell_strings(y_cells, x_cells, self.dim_om)
        # and get Counter dict of the cell strings
        counts = C(cells)

        # use the itemgetter attribute to get the count for each cell
        grid_counts = self.grid_counter(counts)
        # reshape them into an ndarray
        grid_counts = np.reshape(grid_counts, self.gi.shape)
        # and divide the array values by the appropriate
        # grid-cell areas to get the densities
        grid_dens = grid_counts/self.areas

        return grid_dens


class _DensityGridStack:
    def __init__(self, land, window_width=None):

        # dimensions
        self.dim = land.dim

        # resolution (i.e. cell-size)
        self.res = land.res

        # Get the integer closest to 1/10th of the larger of the Landscape dims
        if window_width is None:
            window_width = round(0.1*max(self.dim))

        # set window width of the grid-cell windows within which
        # to count the species' individuals
        self.window_width = window_width

        # get the order of magnitude of the larger of the two land
        # dimensions (used to zero-pad the cell-coordinate strings)
        self.dim_om = land._dim_om

        # get meshgrids of the j and i cell-center coordinates of the
        # landscape-raster cells (to be interpolated to for density
        # calculation)
        self.land_gj, self.land_gi = np.meshgrid(np.arange(0,
                                                           self.dim[0])+0.5,
                                                 np.arange(0, self.dim[1])+0.5)

        # create inner and outer density grids from the land and window-width
        self.grids = dict([(n, g) for n, g in enumerate(
            _make_density_grids(land, self.window_width))])

    def _calc_density(self, x, y):
        # get a concatenated list of the grid-cell center coordinates
        # from all density grids
        pts = np.vstack([self.grids[n].grid_coords for n in range(len(
            self.grids))])
        # and a concatenated list of the densities calculated for
        # all density grids
        vals = np.hstack([self.grids[n]._calc_density(
            x, y).flatten() for n in range(len(self.grids))])

        # then interpolate from those points and values to the centerpoints
        # of all of the land centerpoints
        dens = interpolate.griddata(pts, vals, (self.land_gi, self.land_gj),
                                    method='cubic')
        return dens


class _ConductanceSurface:
    def __init__(self, cond_lyr, mixture, approx_len=5000,
                 vm_distr_kappa=12):
        # dimensions
        self.dim = cond_lyr.dim
        # resolution (i.e. cell-size); defaults to 1
        self.res = cond_lyr.res
        # save whether it uses VonMises mixture dists or not
        self.mix = mixture
        # layer number
        self.lyr_num = cond_lyr.idx
        # set the default approx_len and kappa values if
        # they're accidentally fed in as None values in the params,
        # otherwise set them to the values provided
        if approx_len is None:
            self.approx_len = 5000
        else:
            self.approx_len = approx_len
        if vm_distr_kappa is None:
            self.kappa = 12
        else:
            self.kappa = vm_distr_kappa
        # create the surface
        self.surf = _make_conductance_surface(cond_lyr.rast,
                                              mixture=self.mix,
                                              vm_distr_kappa=self.kappa,
                                              approx_len=self.approx_len)

        assert self.approx_len == self.surf.shape[2], ("_ConductanceSurface."
                                                       "approx_len not equal "
                                                       "to _Conductance"
                                                       "Surface.surf.shape[2]")

    def _draw_directions(self, x, y):
        choices = r.randint(low=0, high=self.approx_len, size=len(x))
        return self.surf[y, x, choices]


class _KDTree:
    def __init__(self, coords, leafsize=100):
        self.tree = cKDTree(data=coords, leafsize=leafsize)

    def _get_mating_pairs(self, coords, dist, choose_nearest=False,
                          inverse_dist_mating=False):
        # just get nearest neighbor pairs, if choose_nearest is requested
        if choose_nearest:
            dists, pairs = self.tree.query(x=coords, k=2,
                                           distance_upper_bound=dist)
            # NOTE: get rid of 'pairs' with no nearest neighbors
            # within the mating radius (they wind up having a dist of np.inf,
            # paired with an index number equal to the last valid index in the
            # KDTree + 1)
            valid_pairs = ~np.isinf(dists[:,1]) 
            dists = dists[valid_pairs, :]
            pairs = pairs[valid_pairs, :]
        # otherwise, get all individuals within mating radius, and get their
        # distances, then make probabilistic draw of each pair using
        # inverse-distance weighting if requested
        else:
            # implement inverse distance-weighted mating, if requested
            if inverse_dist_mating:
                mate_options_cts = {i: l for i, l in enumerate(
                    self.tree.query_ball_point(x=coords, r=dist,
                                               return_length=True)) if l > 1}
                pairs = []
                for focal_ind, ct in mate_options_cts.items():
                    # get the pairwise dists between the focal individ and every
                    # individual who could be a mating option
                    dists_mateopts =self.tree.query(self.tree.data[focal_ind,:],
                                                    k=ct+1,
                                                    distance_upper_bound=dist)
                    valids = (~np.isinf(dists_mateopts[0]) *
                              (dists_mateopts[0] != 0))
                    dists, mateopts = [arr[valids] for arr in dists_mateopts]
                    # calculate probs using linear (not square!)
                    # inverse-dist weighting 
                    probs = (dist - dists)/np.sum((dist-dists))
                    # choose the focal individual's mate
                    chosen_mate = np.random.choice(mateopts, p=probs)
                    # add the chosen pair and their distance to the lists
                    pairs.append([focal_ind, chosen_mate])
            # otherwise, randomly choose an individual within the mating radius
            else:
                mating_options = {i: list(set(l).difference({i})) for i,
                                  l in enumerate(self.tree.query_ball_point(
                                  x=coords, r=dist)) if len(l) > 1}
                # get all individuals within each individual's mating radius
                # NOTE: dropping all individuals who have no valid neighbors
                #       (i.e. whose mating-options lists only include themselves,
                #       and thus are only of length 1)
                # NOTE: the set-diff operation gets rid of the focal_ind as its own
                #       potential mate
                pairs = [[focal_ind, np.random.choice(opts)] for focal_ind,
                         opts in mating_options.items()]
            pairs = np.array(pairs)

        return pairs


# #####################################
# -----------------------------------#
# FUNCTIONS -------------------------#
# -----------------------------------#
# #####################################

# create strings from input cell coordinates
def _make_cell_strings(gi, gj, dim_om):
    # get strings for both i and j cooridnates, zfilling to the
    # correct order-of-magnitude (of the larger land dimensions)
    i_strs = [str(int(i)).zfill(dim_om) for i in gi.flatten()]
    j_strs = [str(int(j)).zfill(dim_om) for j in gj.flatten()]

    # join those strings to one single string, unique for each cell
    cells = [''.join(c) for c in list(zip(i_strs, j_strs))]

    return cells


# make a density grid, based on the Landscape object, the chosen window-width,
# and the Boolean arguments dictating whether or not the grid's x- and y-
# dimension cells should be centered on the land edges (i.e. 0 and dim[_])
def _make_density_grid(land, ww, x_edge, y_edge):

    # half-window width
    hww = ww/2.

    # get land dimensions
    dim = land.dim
    dim_om = land._dim_om

    # create a dictionary of cell ranges, one for when cells center
    # on edge values (i.e. 0 and dim[n] for either dimension),
    # the other for when they don't (i.e. run from hww to dim[n] - hww)
    x_edge_range_dict = {True:  np.arange(0, dim[0]+ww, ww),
                         False: np.arange(0+hww, dim[0]+hww, ww)}
    y_edge_range_dict = {True:  np.arange(0, dim[1]+ww, ww),
                         False: np.arange(0+hww, dim[1]+hww, ww)}

    # create the meshgrid of the centerpoints of neighborhoods (or cells)
    # within which species will be counted
    # (x_edge and y_edge arguments determine whether this grid's
    # x and y cells are centered on the landscape edges or not)
    # NOTE: these are expressed as points in continuous space from 0 to
    # each land dimension, NOT as cell numbers (which will be calculated below)
    gj, gi = np.meshgrid(x_edge_range_dict[x_edge], y_edge_range_dict[y_edge])

    # and get flattened lists of the grid's i and j coordinates
    j = gj.flatten()
    i = gi.flatten()

    # create a single, large Polygon object of the landscape quadrilateral
    land_poly_coords = ((0, 0), (dim[0], 0), (dim[0], dim[1]), (0, dim[1]))
    land_poly = g.Polygon(land_poly_coords)

    # create a list of quadrilaterals centered on each of the points
    # in the grid
    polys = [g.Polygon(((j[n]-hww, i[n]-hww), (j[n]-hww, i[n]+hww),
                        (j[n]+hww, i[n]+hww), (j[n]+hww, i[n]-hww))
                       ) for n in range(len(j))]

    # use the Polygons' intersection method to calculate the total area
    # of the land that is covered by each grid cell (which will be
    # used as the denominator for calculating neighborhood species densities
    # from neighborhood species counts)
    areas = np.reshape([p.intersection(land_poly).area for p in polys],
                       gj.shape)
    # areas of 0 create infinities in divsion, which breaks interpolation
    # and fails to return a density surface, so replace them with very small
    # values, which seems to completely fix the issue and return realistic
    # values for interpolated density surfaces
    areas[areas == 0] = 0.0001

    # get lists of the integer-identifiers (i,j in the proper matrix sense)
    # of the cells in each of the grids (so that these can be matched up
    # with individuals' cell numbers when calculating species density while
    # the model is running)
    # (e.g. a cell centered at point y = 0, x = 5 would be cell i = 0,j = 6
    # if the window-width is 1, or cell i =0,j = 1 if the window width is 5)
    i_cells = (i - (hww * (y_edge))) // ww + (y_edge)
    j_cells = (j - (hww * (x_edge))) // ww + (x_edge)

    # turn the cell-number integers into cell strings
    # (NOTE: the previous algorithm used to calculate species density
    # was more or less the same, but found individuals' cells by
    # checking numerically whether they were within each window, but
    # was much slower and scaled poorly with decreasing window width,
    # increasing landscape size, and increasing population size.
    # This approach instead uses the floor-divide // on individuals'
    # x and y coordinates to generate cell-number integers for them,
    # converts those to strings, then counts the number of instances of each
    # string for the species and uses those values as the counts
    # of individuals within each density-grid cell,
    # obviating the need to loop across grid dimensions.
    # This performs better and scales much better.
    cells = _make_cell_strings(i_cells, j_cells, dim_om)

    # use the above-created data structures to create two DensityGrid objects
    # (which will inhere to the Landscape object as attributes)
    grid = _DensityGrid(dim, dim_om, ww, gi, gj, cells, areas,
                        x_edge=x_edge, y_edge=y_edge)
    return grid


# create 4 density grids, one for each offset (i.e. each
# combination of offset by 0 and by 0.5*window_width)
def _make_density_grids(land, ww):
    # make a grid for each combo of Booleans for x_edge and y_edge
    g1 = _make_density_grid(land, ww, x_edge=True, y_edge=True)
    g2 = _make_density_grid(land, ww, x_edge=False, y_edge=False)
    g3 = _make_density_grid(land, ww, x_edge=True, y_edge=False)
    g4 = _make_density_grid(land, ww, x_edge=False, y_edge=True)
    return(g1, g2, g3, g4)


# Function to generate an approximated Von Mises unimodal distribution
# sampler function
def _make_von_mises_unimodal_sampler(neigh, dirs, vm_distr_kappa=12,
                                     approx_len=5000):
    """Return a list of values approximating a unimodal von Mises distribution
       pointing in the direction of the maximum-valued cell."""
    # list out the directons and their values
    d = list(dirs.ravel())
    n = list(neigh.ravel())
    # get rid of focal cell's values
    del d[4]
    del n[4]
    # identify the direction associated with the maximum-valued cell
    loc = [dirx for idx, dirx in enumerate(d) if n[idx] == max(n)]
    # get mean of max-valued directions, if there are more than 1
    if len(loc) > 1:
        loc = np.mean(loc)
    else:
        loc = loc[0]
    # make the draws for the approximation vector
    approx = s_vonmises.rvs(vm_distr_kappa, loc=loc, scale=1, size=approx_len)
    return approx


# Function to generate a approximated Von Mises mixture distribution
# sampler function
def _make_von_mises_multimodal_sampler(neigh, dirs, vm_distr_kappa=12,
                                       approx_len=5000):
    # Returns a lambda function that is a quick and reliable way to simulate
    # draws from a Von Mises mixture distribution:
    # 1.) Chooses a direction by neighborhood-weighted probability
    # 2.) Makes a random draw from a Von Mises dist centered on the direction,
    # with a vm_distr_kappa value set such that the net effect,
    # when doing this a ton of times for a given neighborhood and then
    # plotting the resulting histogram, gives the visually/heuristically
    # satisfying approximation of a Von Mises mixture distribution

    # NOTE: Just visually, heuristically, vm_distr_kappa = 10 seemed
    # like a perfect middle value (vm_distr_kappa ~3 gives too
    # wide of a Von Mises variance and just chooses values around
    # the entire circle regardless of neighborhood
    # probability, whereas vm_distr_kappa ~20 produces noticeable reductions
    # in probability of moving to directions between the 8 queen's
    # neighborhood directions (i.e. doesn't simulate the mixing well enough)
    # and would generate artefactual movement behavior); 12 also
    # seemed to really well in generating probability valleys
    # when tested on neighborhoods that should generate bimodal distributions
    d = list(dirs.ravel())
    n = list(neigh.ravel())
    del d[4]
    del n[4]
    sum_n = float(sum(n))
    if sum_n > 0:
        n_probs = [i / sum_n for i in n]
    else:
        n_probs = [.125]*8
    loc_choices = r.choice(d, approx_len, replace=True, p=n_probs)
    loc_choices = list(C(loc_choices).items())
    approx = np.hstack([s_vonmises.rvs(vm_distr_kappa, loc=loc, scale=1,
                                       size=size
                                       ) for loc, size in loc_choices])
    return approx


# Runs the von Mises mixture sampler function
# (_make_von_mises_multimodal_sampler) or the
# von Mises unimodal sampler function (make_von_mises_unimodal_sampler)
# across the entire landscape and returns an array-like (list of lists) of the
# resulting lambda-function samplers
def _make_conductance_surface(rast, mixture=True, approx_len=5000,
                              vm_distr_kappa=12):
    queen_dirs = np.array([[-3 * pi / 4, -pi / 2, -pi / 4], [pi, np.NaN, 0],
                           [3 * pi / 4, pi / 2, pi / 4]])

    # copy the landscape raster
    rast = deepcopy(rast)

    # create embedded raster (so that the edge probabilities are
    # appropriately calculated)
    embedded_rast = np.zeros(shape=[n + 2 for n in rast.shape])
    embedded_rast[1:embedded_rast.shape[0] - 1,
                  1:embedded_rast.shape[1] - 1] = rast

    # create a numpy array and store vectors approximating the functions!
    cond_surf = np.float16(np.zeros((rast.shape[0],
                                     rast.shape[1], approx_len)))

    for i in range(rast.shape[0]):
        for j in range(rast.shape[1]):
            neigh = embedded_rast[i:i + 3, j:j + 3].copy()
            if mixture:
                cond_surf[i, j, :] = _make_von_mises_multimodal_sampler(
                    neigh, queen_dirs, vm_distr_kappa=vm_distr_kappa,
                    approx_len=approx_len)
            else:
                cond_surf[i, j, :] = _make_von_mises_unimodal_sampler(
                    neigh, queen_dirs, vm_distr_kappa=vm_distr_kappa,
                    approx_len=approx_len)
    return cond_surf


# coarse wrapper around the nlmpy package
def _make_nlmpy_raster(nlmpy_params):
    if with_nlmpy:
        # pop out the name of the function to be called
        fn_name = nlmpy_params.pop('function')
        # try to create the nlmpy raster
        try:
            # get the function to be called
            fn = getattr(nlmpy, fn_name)
            nlm = fn(**nlmpy_params)
        except Exception as e:
            raise ValueError(("NLMpy could not generate the raster using the "
                              "parameters provided. It threw the following "
                              "error:\n\n\t" "%s\n\n.") % e)
        # if the nlm generated is not constrained between 0 and 1, rescale
        # to that range
        if nlm.min() < 0 or nlm.max() > 1:
            nlm, min_inval, max_inval = _scale_raster(nlm)
    else:
        print("Module 'nlmpy' not installed. Install to use.")
    return nlm


# linearly scale a raster to 0 <= x <= 1, and return the min and max input vals
# as well (for possible reversion)
def _scale_raster(rast, min_inval=None, max_inval=None, min_outval=0,
                  max_outval=1):
    if min_inval is None:
        min_inval = rast.min()
    if max_inval is None:
        max_inval = rast.max()
    scale_rast = (rast - min_inval)/(max_inval - min_inval)
    return(scale_rast, min_inval, max_inval)



#!/usr/bin/python
# viz.py

'''
Defines core functions for visualization
'''

import numpy as np
# import numpy.random as r
# import random
import matplotlib as mpl
import os
import copy
import matplotlib as mpl
if os.environ.get('DISPLAY','') == '':
    print('\n\nNOTE: No display found. Using non-interactive Agg backend\n')
    mpl.use('Agg')
import matplotlib.pyplot as plt
from matplotlib.colors import LinearSegmentedColormap
# from collections import Counter as C
# from operator import itemgetter as ig
# from shapely import geometry as g
# from operator import itemgetter
# from operator import attrgetter
# import sys


######################################
# -----------------------------------#
# FUNCTIONS -------------------------#
# -----------------------------------#
######################################

def _check_display():
    if os.environ.get('DISPLAY','') == '':
        mpl.use('Agg')

def _choose_cmap(lyr_num):
    cols = {0: 'coolwarm',
            1: 'BrBG_r',
            2: 'PRGn',
            3: 'PiYG_r',
            4: 'PuOr_r',
            }
    col = cols[lyr_num % len(cols)]
    return col


def _plot_rasters(land, lyr_num=None, cbar=True, cmap=None, plt_lims=None,
                  vmin=None, vmax=None, lyr_name=None, ticks=None,
                  mask_rast=None,  int_coords=False):
    # if a figure is already open, force colorbar to False,
    # unless explicity force plotted (using 'force' as the argument)
    if plt.get_fignums() and plt.gcf().get_axes() and cbar != 'force':
        cbar = False

    # create a list to hold all colobar ticks
    # (and their min/max vals on the 0-1 raster),
    # and a list to hold all colorbar labels, if cbar == True
    lyr_cbar_ticks = []
    lyr_cbar_labs = []
    # if just a numpy.ndarray or a Layer (not a Landscape object) is
    # provided, or if just a single raster is desired, grab
    # the raster into a list
    if isinstance(land, np.ndarray):
        rasters = [land]
        if lyr_name is not None:
            lyr_names = [lyr_name]
        else:
            lyr_names = ['n/a']
        if lyr_num is not None:
            cmaps = [_choose_cmap(lyr_num)]
        else:
            cmaps = [_choose_cmap(0)]
        lyr_type = ''

    # elif isinstance(land, gnx.landscape.Layer):
    elif 'Layer' in str(type(land)):
        if land.type == 'file':
            rasters = [land._get_rast_in_native_units()]
        else:
            rasters = [land.rast]
        lyr_names = [land.name]
        cmaps = [_choose_cmap(land.idx)]
        lyr_type = land.type
        if cbar in (True, 'force'):
            lyr_cbar_ticks.append(land._get_cbar_ticks_and_minmax_scaled_vals(
                                                                             ))
            lyr_cbar_labs.append(land.units)
        # get the cell bounds
        x_cell_bds, y_cell_bds = land._x_cell_bds, land._y_cell_bds
    # elif isinstance(land, gnx.landscape.Landscape):
    elif 'Landscape' in str(type(land)):
        if lyr_num is not None:
            if land[lyr_num].type == 'file':
                rasters = [land[lyr_num]._get_rast_in_native_units()]
            else:
                rasters = [land[lyr_num].rast]
            lyr_names = [land[lyr_num].name]
            cmaps = [_choose_cmap(lyr_num)]
            lyr_cbar_ticks.append(
                        land[lyr_num]._get_cbar_ticks_and_minmax_scaled_vals())
            lyr_cbar_labs.append(land[lyr_num].units)
        # else just create a list of all rasters
        else:
            rasters = [lyr.rast if (
                lyr.type != 'file') else lyr._get_rast_in_native_units(
                                                    ) for lyr in land.values()]
            lyr_names = [lyr.name for lyr in land.values()]
            cmaps = [_choose_cmap(lyr.idx) for lyr in land.values()]
            [lyr_cbar_ticks.append(
                land[lyr_num]._get_cbar_ticks_and_minmax_scaled_vals(
                                                    )) for lyr_num in [*land]]
            [lyr_cbar_labs.append(land[lyr_num].units) for lyr_num in [*land]]
        # else just create a list of all rasters
        lyr_types = [lyr.type for lyr in land.values()]
        lyr_type = 'file' * ('file' in lyr_types)
        # get the cell bounds
        x_cell_bds, y_cell_bds = land._x_cell_bds, land._y_cell_bds
    # if integer-coordinates asked for, or land is a numpy array,
    # get the cell-bounds from cell-number integers
    if isinstance(land, np.ndarray) or int_coords:
        y_cell_bds, x_cell_bds = [np.linspace(0, dim,
                                              dim+1) for dim in land.shape]

    # plot all with the same cmap, if the cmap argument was provided
    if isinstance(cmap, str):
        # get the requested cmap
        cmap = getattr(plt.cm, cmap)
        cmaps = [cmap] * len(rasters)

    # mask all arrays, and set cmaps' bad values to gray,
    # if mask_rast is provided
    if mask_rast is not None:
        rasters = [np.ma.masked_where(np.isnan(mask_rast),
                                      rast) for rast in rasters]
        cmaps = [copy.copy(getattr(plt.cm, cm)) for cm in cmaps]
        [cm.set_bad('#8C8C8C') for cm in cmaps]

    # create alphas list
    alphas = [1] + [0.5] * (len(rasters)-1)
    # create vmin and vmax lists, if None
    if (vmin is None
       or isinstance(vmin, float)
        # DEH: 10-28-19: For some reason isinstance(<np.float32 obj>, float)
        # returns False, so check if np.floating to avoid error
       or isinstance(vmin, np.floating)
       or isinstance(vmin, int)):
        vmin = [vmin] * len(rasters)
    if (vmax is None
       or isinstance(vmax, float)
       or isinstance(vmax, np.floating)
       or isinstance(vmax, int)):
        vmax = [vmax] * len(rasters)
    # plot all the rasters...
    for n in range(len(rasters)):
        # pull out the zoomed raster, if requested
        # if zoom is not None:
        #    min_i, max_i = zoom[0]
        #    min_j, max_j = zoom[1]
        #    rasters[n] = np.array([row[
                    # min_j:max_j] for row in rasters[n][min_i:max_i]])
        plt.pcolormesh(x_cell_bds, y_cell_bds, rasters[n], cmap=cmaps[n],
                       vmin=vmin[n], vmax=vmax[n], alpha=alphas[n])
        plt.axis('scaled')
        if ((lyr_type != 'file' and not ticks) or
           (lyr_type == 'file' and ticks is False)):
            plt.xticks([])
            plt.yticks([])
        else:
            # add geo-coordinate ticks, if this is a raster from a file
            if lyr_type == 'file':
                #    (x_ticks, x_tick_labs,
                #     y_ticks, y_tick_labs) = land._get_coord_ticks()
                    # format at ticklabels to the same decimal place
                #    float_fmt = ('%0.' + str(max([len(str(n).split(
                #         '.')[1]) for n in x_tick_labs + y_tick_labs])) + 'f')
                #    x_tick_labs = [float_fmt % n for n in x_tick_labs]
                #    y_tick_labs = [float_fmt % n for n in y_tick_labs]
                #    plt.xticks(x_ticks, x_tick_labs, rotation=90)
                #    plt.yticks(y_ticks, y_tick_labs)
                #    ax = plt.gca()
                plt.xlabel('lon')
                plt.ylabel('lat')
        if plt_lims is not None:
            plt.xlim(plt_lims[0])
            plt.ylim(plt_lims[1])
        # and their colorbars, if requested
        if cbar:
            if len(lyr_cbar_ticks) == 0:
                # cbar_max_bound = max(rasters[n].max(),
                #                     [1 if vmax[n] is None else vmax[n]][0])
                # cbar_min_bound = min(rasters[n].min(),
                #                     [0 if vmin[n] is None else vmin[n]][0])
                cbar_min_bound = vmin[n]
                cbar_max_bound = vmax[n]
                if cbar_min_bound is None and cbar_max_bound is None:
                    cbar_min_bound = 0
                    cbar_max_bound = 1
                cbar_bounds = np.linspace(cbar_min_bound, cbar_max_bound, 51)
                cbar = plt.colorbar(boundaries=cbar_bounds)
            else:
                cbar = plt.colorbar(ticks=np.linspace(lyr_cbar_ticks[n][1],
                                                      lyr_cbar_ticks[n][2], 5))
                cbar.ax.set_yticklabels(lyr_cbar_ticks[n][0])
                cbar.set_label(lyr_cbar_labs[n], rotation=270,
                               labelpad=15, y=0.5)
            cbar.ax.set_title("layer: %s" % lyr_names[n])
            title = cbar.ax.title
            font = mpl.font_manager.FontProperties(family='sans-serif',
                                                   style='normal', size=11)
            title.set_font_properties(font)


def _plot_points(points, lyr_num=None, color='black',
                 edge_color='face', text_color='black', linewidth=0.5,
                 pt_cmap=None, size=25, text_size=9, alpha=False, text=None,
                 plt_lims=None, vmin=None, vmax=None, animate=False):
    #get the x and y coordinates from the points (and subtract 0.5
    #to line the points up with the plt.imshow() grid of a
    #landscape raster; imshow plots each pixel centered on its 
    #index, but the points then plot on those indices, so wind up
    #shifted +0.5 on each axis
    x = points[:, 0]
    y = points[:, 1]
    #handle the alpha value as necessary
    if alpha == True and type(alpha) == bool:
        alpha = 0.6
    elif alpha != False and type(alpha) in (int, float):
        assert alpha >= 0 and alpha <= 1, ("Values of 'alpha' must be between "
            "0 and 1.")
        alpha = alpha
    else:
        alpha = 1.0

    #plot the points, as stipulated by arguments
    if pt_cmap is not None:
        if pt_cmap == 'terrain':
            colors = ['#3C22B4', '#80A6FF', '#FFFFFF']
            # colors to match matplotlib colormap 'terrain' palette
            #extremes, but with hybrid a mix of the extremes
            # rather than the yellow at the middle of the palette,
            #for nicer viewing
            cmap = LinearSegmentedColormap.from_list('my_cmap', colors, N=50)
        elif type(pt_cmap) == str:
            cmap = getattr(plt.cm, pt_cmap)
        elif isinstance(pt_cmap, LinearSegmentedColormap):
            cmap = pt_cmap
        points = plt.scatter(x, y, s=size, c=color, cmap=cmap,
                             linewidth=linewidth, edgecolor=edge_color,
                             alpha=alpha, vmin=vmin, vmax=vmax)
    else:
        points = plt.scatter(x, y, s=size, c=color, linewidth=linewidth,
                             edgecolor=edge_color, alpha=alpha, vmin=vmin,
                             vmax=vmax)

    #add text, if requested
    if text is not None:
        plot_text = []
        for n,t in enumerate(text):
            if plt_lims is not None:
                if (plt_lims[0][0] <= x[n] <= plt_lims[0][1]
                    and plt_lims[1][0] <= y[n] <= plt_lims[1][1]):
                    plot_text.append((x[n], y[n], t))
            else:
                plot_text.append((x[n], y[n], t))
        [plt.text(*item, color=text_color, size=text_size,
                                        alpha=alpha) for item in plot_text];

    if (plt_lims is not None
        and len(plt_lims) == 2
        and [len(item) for item in plt_lims] == [2,2]):
        plt.xlim(plt_lims[0])
        plt.ylim(plt_lims[1])
    else:
        print(("plt_lims appears not to be a valid argument "
               "(i.e. a 2-tuple of 2-tuples)"))

    #overwrite points with None, unless animate == True
    if not animate:
        points = None

    return points


def _get_lyr_plt_lims(land):
    # NOTE: these are set up so that 0,0 is in the upper-left corner
    xlim, ylim = [tuple(np.sort((land.ulc[i] - 0.5 * land.res[i],
                                 land.ulc[i] + (land.dim[i] + 0.5) * land.res[
                                                       i]))) for i in range(2)]
    lims = (xlim, ylim)
    return(lims)


def _get_zoom_plt_lims(x, y, zoom_width):
    # get zoom-half-width
    zhw = zoom_width/2
    xlim = (x - zhw, x + zhw)
    ylim = (y + zhw, y - zhw)
    lims = (xlim, ylim)
    return(lims)


def _get_plt_lims(land=None, x=None, y=None, zoom_width=None):
    if zoom_width is not None and x is not None and y is not None:
        plt_lims = _get_zoom_plt_lims(x, y, zoom_width)
    else:
        plt_lims = _get_lyr_plt_lims(land)
    return(plt_lims)


def _make_fitness_cmap_and_cbar_maker(min_val, max_val = 1,
                        cmap = 'gray', max_cmap_len = 5, trt_num = None):
    # define the colormap
    cmap = getattr(plt.cm, cmap)
    #extract all the colors into a list
    cmap_list = [cmap(i) for i in range(cmap.N)]
    #create new list, with the majority of the color range expressed for
    #the values between 1 and the min_val, then the remainder stretched
    #out between min_val and 0
    #top = np.int64(np.linspace(0,len(cmap_list)*0.15,max_cmap_len*0.8))
    #bot = np.int64(np.linspace(1+(len(cmap_list)*0.15),
                                    #len(cmap_list)-1, max_cmap_len*0.2))
    new_cmap_inds = np.int64(np.linspace(0, len(cmap_list), max_cmap_len))
    #new_cmap_inds = list(np.hstack((top,bot)))
    new_cmap_inds = list(set(new_cmap_inds))
    new_cmap_list = [col for n,col in enumerate(
                                            cmap_list) if n in new_cmap_inds]
    # create the new map
    #cmap = cmap.from_list('Custom cmap', new_cmap_list, len(cmap_list))
    # define the bin-boundaries 
    #lower_bounds = np.linspace(0,min_val,round((2*cmap.N/3)+1))[:-1]
    #upper_bounds = np.linspace(min_val, max_val,round(cmap.N/3))
    #bounds = np.hstack((lower_bounds, upper_bounds))
    bounds = np.linspace(min_val, max_val, cmap.N)
    assert len(bounds) == cmap.N
    #normalize the colormap
    #norm = mpl.colors.BoundaryNorm(bounds, cmap.N)
    #create ticks for the colorbar
    ticks_inds = np.int64(np.linspace(0, len(bounds)-1, 10))
    ticks = list(bounds[ticks_inds])
    tick_closest_to_min_val = min([abs(tick-min_val) for tick in ticks])
    ind_closest = [n for n,tick in enumerate(ticks) if abs(
                                tick-min_val) == tick_closest_to_min_val][0]
    ticks[ind_closest] = min_val
    ticks = sorted(ticks)
    ticks = [round(tick, 2) for tick in ticks]
    if trt_num is None:
        tick_labs = [(' '*10 + 'min. fit. =\n' + ' ' * 10 + ('$1-\prod_'
                      '{trait=1}^{t} \phi_{t} \prod_{del.mut.=1}'
                      '^{d} \phi_{d}$\n')) if n == ind_closest else str(
                        tick) for n, tick in enumerate(ticks)]
    else:
        tick_labs = [(' ' * 10 + 'min. fit. =\n' + ' ' * 10 + ('$1-\phi_'
                    '{trait=%i}$\n')) % (trt_num) if n == ind_closest else str(
                        tick) for n, tick in enumerate(ticks)]
    #create a function for making the colorbar, to be shipped out to and
    #called within species.Species.plot_fitness()
    def make_cbar(ax):
        cbar = mpl.colorbar.ColorbarBase(ax, cmap=cmap,
            spacing='proportional', ticks=ticks, boundaries=bounds,
            format='%1i')
        cbar.set_ticks(ticks)
        cbar.set_ticklabels(tick_labs)
    return(cmap, make_cbar)

def _make_fitness_cbar(make_cbar_fn, min_fit):
    fig = plt.gcf()
    ax1 = plt.gca()
    ax2 = fig.add_axes([0.84, 0.106, 0.02, 0.7774])
    make_cbar_fn(ax2)
    #ax2.plot([0,1],[round(min_fit,2)]*2, c = 'black', lw = 1)
    ax2.set_title('fitness')
    title = ax2.title
    font = mpl.font_manager.FontProperties(family='sans-serif',
                                                style='normal', size=10)
    title.set_font_properties(font)

# return an arbitrarily lighter version of a color
# (stolen and tweaked from Chase Seibert: https://chase-seibert.github.io/blog/
# 2011/07/29/python-calculate-lighterdarker-rgb-colors.html)
def _calc_reshaded_color(hex_color, brightness_offset=1):
    """ takes a color like #87c95f and produces a lighter or darker variant
    """
    if len(hex_color) != 7:
        raise Exception(("Passed %s into color_variant(), needs to be "
                        "in #87c95f format.") % hex_color)
    rgb_hex = [hex_color[x:x+2] for x in [1, 3, 5]]
    new_rgb_int = [int(hex_value,
                       16) + brightness_offset for hex_value in rgb_hex]
    # make sure new values are between 0 and 255
    new_rgb_int = [min([255, max([0, i])]) for i in new_rgb_int]
    # hex() produces "0x88", we want just "88"
    new_hex_int = [hex(i)[2:] for i in new_rgb_int]
    new_hex_int = [str(i).zfill(2) for i in new_hex_int]
    new_hex = "#" + "".join(new_hex_int)
    return new_hex



#!/usr/bin/python
# _simult_select.py

# flake8: noqa

# geonomics imports
from geonomics.utils.viz import _check_display

# other imports
import os
import numpy as np
import matplotlib as mpl
_check_display()
import matplotlib.pyplot as plt
from matplotlib import gridspec
from mpl_toolkits.axes_grid1 import make_axes_locatable
import time


# define a function to calculate the mean difference between phenotype
# and environment for a species
def calc_mean_z_e_diff(spp, trait_num=0):
    zs = spp._get_z()[:, trait_num].ravel()
    es = spp._get_e(lyr_num=spp.gen_arch.traits[trait_num].lyr_num)
    mean_diff = np.mean(np.abs(zs - es))
    return mean_diff


def _make_params():
    array_center_vals = np.array([*np.linspace(0.25, 0.75, 20)][::-1])
    array_center_vals = 0.5+-0.5*np.sin(2*np.pi*array_center_vals)
    array_vals = np.hstack(([1]*15, array_center_vals, [0]*15))
    env_array0 = np.vstack([array_vals]*50)
    env_array1= env_array0.T

    K_array = np.ones((50,50))

    #This is a default parameters file generated by Geonomics
    #(by the gnx.params.make_parameters_file() function).


                          ## :: ::    :::            ##
                    ##:::   :::::    :::   ::    :: :: :::##
                 ## ::::     ::           ::   ::::::::::::::##
               ##::::::                       ::::::::: :::::: :##
             ## :    :::                    :::    ::    :::::::::##
            ##GGGGG EEEE OOO   N   N   OOO   M   M IIIII  CCCC SSSS##
           ##G     E    O   O  NN  N  O   O  M   M   I   C     S    ##
           ##G     EEE O     O N N N O     O MM MM   I   C     SSSSS##
           ##G GGG EEE O     O N  NN O     O M M M   I   C         S##
           ##G   G E    O   O  N   N  O   O  M   M   I   C        SS##
            ##GGGG  EEEE OOO   N   N   OOO   M   M IIIII  CCCC SSSS##
             ##  ::::::::        ::::::::::::  :       ::  ::   : ##
               ##  ::::              :::::::  ::     ::::::::  :##
                 ## :::               :::::: ::       ::::::  ##
                    ##:                ::::                ##
                          ##                         ##



    params = {
    ###############################################################################

    ###################
    #### LANDSCAPE ####
    ###################
        'landscape': {

        ##############
        #### main ####
        ##############
            'main': {
                #y,x (a.k.a. i,j) dimensions of the Landscape
                'dim':                      (50,50),
                #resolution of the Landscape
                'res':                      (1,1),
                #upper-left corner of the Landscape
                'ulc':                      (0,0),
                #projection of the Landscape
                'prj':                      None,
                }, # <END> 'main'

        ################
        #### layers ####
        ################
            'layers': {

                #layer name (LAYER NAMES MUST BE UNIQUE!)
                'layer_0': {

            #######################################
            #### layer num. 0: init parameters ####
            #######################################

                    #initiating parameters for this layer
                    'init': {

                        #parameters for a 'defined'-type Layer
                        'defined': {
                            #raster to use for the Layer
                            'rast':                 env_array0,
                            #point coordinates
                            'pts':                  None,
                            #point values
                            'vals':                 None,
                            #interpolation method {None, 'linear', 'cubic',
                            #'nearest'}
                            'interp_method':        None,

                            }, # <END> 'defined'

                        }, # <END> 'init'

                    }, # <END> layer num. 0


                #layer name (LAYER NAMES MUST BE UNIQUE!)
                'layer_1': {

            #######################################
            #### layer num. 1: init parameters ####
            #######################################

                    #initiating parameters for this layer
                    'init': {

                        #parameters for a 'defined'-type Layer
                        'defined': {
                            #raster to use for the Layer
                            'rast':                 env_array1,
                            #point coordinates
                            'pts':                  None,
                            #point values
                            'vals':                 None,
                            #interpolation method {None, 'linear', 'cubic',
                            #'nearest'}
                            'interp_method':        None,

                            }, # <END> 'defined'

                        }, # <END> 'init'

                    }, # <END> layer num. 1



                #layer name (LAYER NAMES MUST BE UNIQUE!)
                'layer_2': {

            #######################################
            #### layer num. 1: init parameters ####
            #######################################

                    #initiating parameters for this layer
                    'init': {

                        #parameters for a 'defined'-type Layer
                        'defined': {
                            #raster to use for the Layer
                            'rast':                 K_array,
                            #point coordinates
                            'pts':                  None,
                            #point values
                            'vals':                 None,
                            #interpolation method {None, 'linear', 'cubic',
                            #'nearest'}
                            'interp_method':        None,

                            }, # <END> 'defined'

                        }, # <END> 'init'

                    }, # <END> layer num. 2


        #### NOTE: Individual Layers' sections can be copy-and-pasted (and
        #### assigned distinct keys and names), to create additional Layers.


                } # <END> 'layers'

            }, # <END> 'landscape'


    ###############################################################################

    ###################
    #### COMMUNITY ####
    ###################
        'comm': {

            'species': {

                #species name (SPECIES NAMES MUST BE UNIQUE!)
                'spp_0': {

                #####################################
                #### spp num. 0: init parameters ####
                #####################################

                    'init': {
                        #starting number of individs
                        'N':                1000,
                        #carrying-capacity Layer name
                        'K_layer':          'layer_2',
                        #multiplicative factor for carrying-capacity layer
                        'K_factor':         0.5,
                        }, # <END> 'init'

                #######################################
                #### spp num. 0: mating parameters ####
                #######################################

                    'mating'    : {
                        #age(s) at sexual maturity (if tuple, female first)
                        'repro_age':                0,
                        #whether to assign sexes
                        'sex':                      False,
                        #ratio of males to females
                        'sex_ratio':                1/1,
                        #intrinsic growth rate
                        'R':                        0.5,
                        #intrinsic birth rate (MUST BE 0<=b<=1)
                        'b':                        0.5,
                        #expectation of distr of n offspring per mating pair
                        'n_births_distr_lambda':    1,
                        #whether n births should be fixed at n_births_dist_lambda
                        'n_births_fixed':           True,
                        #radius of mate-search area
                        'mating_radius':            3,
                        'choose_nearest_mate':      False,
                        'inverse_dist_mating':      False,
                        }, # <END> 'mating'

                ##########################################
                #### spp num. 0: mortality parameters ####
                ##########################################

                    'mortality'     : {
                        #maximum age
                        'max_age':                      5,
                        #min P(death) (MUST BE 0<=d_min<=1)
                        'd_min':                        0,
                        #max P(death) (MUST BE 0<=d_max<=1)
                        'd_max':                        1,
                        #width of window used to estimate local pop density
                        'density_grid_window_width':    None,
                        }, # <END> 'mortality'

                #########################################
                #### spp num. 0: movement parameters ####
                #########################################

                    'movement': {
                        #whether or not the species is mobile
                        'move':                     True,
                        #mode of distr of movement direction
                        'direction_distr_mu':       1,
                        #concentration of distr of movement direction
                        'direction_distr_kappa':    0,
                        #mean of distr of movement distance
                        'movement_distance_distr_param1':        0.5,
                        #variance of distr of movement distance
                        'movement_distance_distr_param2':     0.5,
                        'movement_distance_distr':          'wald',
                        #mean of distr of dispersal distance
                        'dispersal_distance_distr_param1':       0.5,
                        #variance of distr of dispersal distance
                        'dispersal_distance_distr_param2':    0.5,
                        'dispersal_distance_distr':     'wald',
                        },    # <END> 'movement'


                #####################################################
                #### spp num. 0: genomic architecture parameters ####
                #####################################################

                    'gen_arch': {
                        #file defining custom genomic arch
                        'gen_arch_file':            None,
                        #num of loci
                        'L':                        1000,
                        #num of chromosomes
                        'l_c':                      [1000],
                        #whether starting allele frequencies should be fixed at 0.5
                        'start_p_fixed':            0.5,
                        #genome-wide per-base neutral mut rate (0 to disable)
                        'mu_neut':                  0,
                        #genome-wide per-base deleterious mut rate (0 to disable)
                        'mu_delet':                 0,
                        #shape of distr of deleterious effect sizes
                        'delet_alpha_distr_shape':  0.2,
                        #scale of distr of deleterious effect sizes
                        'delet_alpha_distr_scale':  0.2,
                        #alpha of distr of recomb rates
                        'r_distr_alpha':            None,
                        #beta of distr of recomb rates
                        'r_distr_beta':             None,
                        #whether loci should be dominant (for allele '1')
                        'dom':                      False,
                        #whether to allow pleiotropy
                        'pleiotropy':               False,
                        #custom fn for drawing recomb rates
                        'recomb_rate_custom_fn':    None,
                        #number of recomb paths to hold in memory
                        'n_recomb_paths_mem':       int(1e4),
                        #total number of recomb paths to simulate
                        'n_recomb_paths_tot':       int(1e5),
                        'n_recomb_sims':            10000,
                        'start_neut_zero':          False,
                        'allow_ad_hoc_recomb':      False,
                        'jitter_breakpoints':       False,
                        #whether to save mutation logs
                        'mut_log':                  False,

                        'traits': {

                            ###########################
                            ####trait 0 parameters ####
                            ###########################
                            #trait name (TRAIT NAMES MUST BE UNIQUE!)
                            'trait_0': {
                                #trait-selection Layer name
                                'layer':                'layer_0',
                                #polygenic selection coefficient
                                'phi':                  0.05,
                                #number of loci underlying trait
                                'n_loci':               10,
                                #mutation rate at loci underlying trait
                                'mu':                   0,
                                #mean of distr of effect sizes
                                'alpha_distr_mu' :      0.1,
                                #variance of distr of effect size
                                'alpha_distr_sigma':    0,
                                #max alpha value
                                'max_alpha_mag':        None,
                                #curvature of fitness function
                                'gamma':                1,
                                #whether the trait is universally advantageous
                                'univ_adv':             False
                                }, # <END> trait 0

                            ###########################
                            ####trait 1 parameters ####
                            ###########################
                            #trait name (TRAIT NAMES MUST BE UNIQUE!)
                            'trait_1': {
                                #trait-selection Layer name
                                'layer':                'layer_1',
                                #polygenic selection coefficient
                                'phi':                  0.05,
                                #number of loci underlying trait
                                'n_loci':               10,
                                #mutation rate at loci underlying trait
                                'mu':                   0,
                                #mean of distr of effect sizes
                                'alpha_distr_mu' :      0.1,
                                #variance of distr of effect size
                                'alpha_distr_sigma':    0,
                                #max alpha value
                                'max_alpha_mag':        None,
                                #curvature of fitness function
                                'gamma':                1,
                                #whether the trait is universally advantageous
                                'univ_adv':             False
                                }, # <END> trait 1

        #### NOTE: Individual Traits' sections can be copy-and-pasted (and
        #### assigned distinct keys and names), to create additional Traits.


                            }, # <END> 'traits'

                        }, # <END> 'gen_arch'


                    }, # <END> spp num. 0



        #### NOTE: individual Species' sections can be copy-and-pasted (and
        #### assigned distinct keys and names), to create additional Species.


                }, # <END> 'species'

            }, # <END> 'comm'


    ###############################################################################

    ###############
    #### MODEL ####
    ###############
        'model': {
            #total Model runtime (in timesteps)
            'T':            1000,
            #min burn-in runtime (in timesteps)
            'burn_T':       60,
            #seed number
            'num':          None,
            #time step interval for simplification of tskit tables
            'tskit_simp_interval':      100,

            ###############################
            #### iterations parameters ####
            ###############################
            'its': {
                #num iterations
                'n_its':            1,
                #whether to randomize Landscape each iteration
                'rand_landscape':   False,
                #whether to randomize Community each iteration
                'rand_comm':        False,
                #whether to burn in each iteration
                'repeat_burn':      True,
                }, # <END> 'iterations'


            ####################################
            #### data-collection parameters ####
            ####################################
            'data': {
                'sampling': {
                    #sampling scheme {'all', 'random', 'point', 'transect'}
                    'scheme':               'random',
                    #sample size at each point, for point & transect sampling
                    'n':                    801,
                    #coords of collection points, for point sampling
                    'points':               None,
                    #coords of transect endpoints, for transect sampling
                    'transect_endpoints':   None,
                    #num points along transect, for transect sampling
                    'n_transect_points':    None,
                    #collection radius around points, for point & transect sampling
                    'radius':               None,
                    #when to collect data
                    'when':                 0,
                    #whether to save current Layers when data is collected
                    'include_landscape':    False,
                    #whether to include fixed loci in VCF files
                    'include_fixed_sites':  False,
                    },
                'format': {
                    #format for genetic data {'vcf', 'fasta'}
                    'gen_format':           ['vcf', 'fasta'],
                    #format for vector geodata {'csv', 'shapefile', 'geojson'}
                    'geo_vect_format':      'csv',
                    #format for raster geodata {'geotiff', 'txt'}
                    'geo_rast_format':      'geotiff',
                    },
                }, #<END> 'data'

            #####################################
            #### stats-collection parameters ####
            #####################################
            'stats': {
                #mean fitness
                'mean_fit': {
                    #whether to calculate
                    'calc': True,
                    #calculation frequency (in timesteps)
                    'freq': 5,
                    },
                 }, # <END> 'stats'

            } # <END> 'model'

        } # <END> params

    return params


def _run(params, save_figs=False, time_it=False):

    # create data structure to store z-e diffs for both traits
    z_e_diffs = {0: [], 1: []}

    # start timer
    if time_it:
        start = time.time()

    # set model name (since not being read in from separate params file)
    params.model['name'] = 'simult_select_demo'

    # make the model
    from .. import make_model
    mod = make_model(params)

    # get total runtime from params
    T = mod.params.model.T

    # run model
    mod.walk(T=10000, mode='burn', verbose=True)
    for t in range(T):
        if not time_it:
            for trt_num in range(len(mod.comm[0].gen_arch.traits)):
                z_e_diffs[trt_num].append(calc_mean_z_e_diff(mod.comm[0],
                                                             trait_num=trt_num))
        mod.walk(T=1, mode='main', verbose=True)
    if not time_it:
        for trt_num in range(len(mod.comm[0].gen_arch.traits)):
            z_e_diffs[trt_num].append(calc_mean_z_e_diff(mod.comm[0],
                                                         trait_num=trt_num))

    # stop timer
    if time_it:
        stop = time.time()
        tot_time = stop - start

    # plot the resulting species on top of each layer
    fig = plt.figure(figsize=(15, 10))
    fig.tight_layout()
    gs = gridspec.GridSpec(1, 2, width_ratios=[1, 1])
    im = plt.pcolormesh(np.linspace(0, mod.land.dim[0], mod.land.dim[0]+1),
                        np.linspace(0, mod.land.dim[1], mod.land.dim[1]+1),
                        mod.land[0].rast, cmap='coolwarm')
    ax1 = plt.subplot(gs[0])
    mod.plot_phenotype(0, 0, 0, size=85)
    divider = make_axes_locatable(ax1)
    cax = divider.append_axes("right", size="5%", pad=0.05)
    cbar = plt.colorbar(im, cax=cax)
    cbar.set_label('environmental and phenotypic value', rotation=270,
                   labelpad=25, y=0.5, size=18)
    ax2 = plt.subplot(gs[1])
    im = plt.pcolormesh(np.linspace(0, mod.land.dim[0], mod.land.dim[0]+1),
                        np.linspace(0, mod.land.dim[1], mod.land.dim[1]+1),
                        mod.land[1].rast, cmap='BrBG_r')
    mod.plot_phenotype(0, 1, 1, size=85)
    divider = make_axes_locatable(ax2)
    cax = divider.append_axes("right", size="5%", pad=0.05)
    cbar = plt.colorbar(im, cax=cax)
    cbar.set_label('environmental and phenotypic value', rotation=270,
                   labelpad=25, y=0.5, size=18)
    ax1.set_title('trait 0', size=20)
    ax2.set_title('trait 1', size=20)
    #axes_list = fig.axes
    #axes[1].tick_params(labelsize=15)
    #axes[3].tick_params(labelsize=15)
    plt.show()

    # print out time
    if time_it:
        print("Model ran in %0.2f seconds." % tot_time)

    # save the figure
    if save_figs:
        plt.savefig('sim_sel.pdf', format='png', dpi=1000)


    # plot z-e diffs
    z_e_fig = plt.figure()
    ax = z_e_fig.add_subplot(111)
    trt_colors = ['#096075', '#e87d4f']
    plt.plot(range(len(z_e_diffs[0])), z_e_diffs[0], trt_colors[0])
    plt.plot(range(len(z_e_diffs[1])), z_e_diffs[1], trt_colors[1])
    ax.legend(labels=['trait = %i' % trt for trt in range(2)],
              loc='best', prop={'size': 18})
    ax.set_xlabel('time', size=18)
    ax.set_ylabel(('mean difference between individuals\'\nphenotypes and '
                   'environmental values'), size=18)
    ax.tick_params(labelsize=15)
    plt.show()
    if save_figs:
        plt.savefig('sim_sel_z-e_plot.png', format='png', dpi=1000)

    return mod


